'''
The copyright in this software is being made available under the Clear BSD
License, included below. No patent rights, trademark rights and/or
other Intellectual Property Rights other than the copyrights concerning
the Software are granted under this license.

The Clear BSD License

Copyright (c) 2019-2025, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The NNCodec Authors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted (subject to the limitations in the disclaimer below) provided that
the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

     * Neither the name of the copyright holder nor the names of its
     contributors may be used to endorse or promote products derived from this
     software without specific prior written permission.

NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
'''

import enum
import numpy as np
import sys
from nncodec.nnc_core import nnr_model

assert sys.byteorder == "little"


class NnrUnitType(enum.IntEnum):
    NNR_STR = 0
    NNR_MPS = 1
    NNR_LPS = 2
    NNR_TPL = 3
    NNR_QNT = 4
    NNR_NDU = 5
    NNR_AGG = 6

    
class DecompressedDataFormat(enum.IntEnum):
    TENSOR_INT32   = 0
    TENSOR_FLOAT32 = 1
    TENSOR_INT2    = 2
    TENSOR_INT3    = 3
    TENSOR_INT4    = 4
    TENSOR_INT8    = 5
    TENSOR_INT16   = 6
    TENSOR_INT64   = 7
    TENSOR_FLOAT16 = 8
    TENSOR_FLOAT64 = 9

    
class CompressedDataUnitPayloadType(enum.IntEnum):
    NNR_PT_INT       = 0 ##tdb not yet implemented
    NNR_PT_FLOAT     = 1
    NNR_PT_RAW_FLOAT = 2
    NNR_PT_BLOCK     = 3

    
class BlockParameterTypes(enum.IntEnum):
    NNR_CPT_DC = 0x01
    NNR_CPT_LS = 0x02
    NNR_CPT_BN = 0x04
    NNR_CPT_BI = 0x08

    
class QuantizationMethodFlags(enum.IntEnum):
    NNR_QSU = 1
    NNR_QCB = 2

class PruningInformationRepresentationTypes(enum.IntEnum):
    NNR_TPL_BMSK = 0
    NNR_TPL_DICT = 1


class ParentNodeIdType(enum.IntEnum):
    ICNN_NDU_ID        = 0
    ICNN_NDU_PL_SHA256 = 1
    ICNN_NDU_PL_SHA512 = 2

class BitWriter():
    def __init__( self, bitstream ):
        self.__byteList = bitstream
        self.__bitPos = -1
        self.__bytePos = -1
    
    def getNumBitsTouched(self):
        return (self.__bytePos + 1) * 8 - self.__bitPos - 1

    def writeBit(self, bit ):
        assert bit == 0 or bit == 1
        if self.__bitPos < 0:
            self.__bytePos += 1
            if len(self.__byteList) <= self.__bytePos:
                self.__byteList.append( 0 )
            self.__bitPos = 7
        if bit:
            self.__byteList[self.__bytePos] |= 1 << self.__bitPos
        else:
            self.__byteList[self.__bytePos] &= 255 - (1 << self.__bitPos)
        self.__bitPos -= 1

    def u(self, n, x):
        assert n > 0
        assert x >= 0 and x < (1<<n)
        for i in range( n-1, -1, -1):
            self.writeBit( (np.uint64(x) >> np.uint64(i)) & np.uint64(1) )

    def ue(self, k, x):
        while x >= (1<<k):
            self.u(1, 0)
            x -= 1<<k
            k +=1

        self.u(1, 1)
        if k > 0:
            self.u(k, x)

    def i(self, n, x):
        assert n > 0
        assert x >= -(1<<(n-1)) and x < (1<<(n-1))
        self.u( n, x if x >= 0 else x + (1<<n) )

    def ie(self, k, x):
        x = ((-x)<<1) if x <= 0 else ((x<<1)-1)
        self.ue(k,x)

    def byte_alignment(self):
        self.u(1,1)
        self.__bitPos = -1
        
    def flt(self, n, x):
        assert n == 32
        assert self.__bytePos + 1 == len(self.__byteList) # only supported at end of the bitstream
        assert isinstance( x, np.float32 )
        for byte in np.float32(x, dtype='<f4').tobytes():
            self.u(8, byte)

    def flt_tensor(self, n, dims, x):
        assert self.__bitPos == -1
        assert n == 32
        assert self.__bytePos + 1 == len(self.__byteList) # only supported at end of the bitstream
        assert isinstance( x, np.ndarray )
        assert x.dtype == np.float32
        self.__byteList.extend( x.tobytes() )
        self.__bytePos = len(self.__byteList) - 1

    def st(self, v):
        assert self.__bitPos == -1
        assert isinstance( v, str )
        assert self.__bytePos + 1 == len(self.__byteList) # only supported at end of the bitstream
        self.__byteList.extend( v.encode( "utf-8", "strict" ) )
        self.__byteList.append( 0 )
        self.__bytePos = len(self.__byteList) - 1

    def codebook(self, codebook_egk, codebook_size, CbZeroOffset, codebook):
        previousValue = codebook[CbZeroOffset]
        self.ie(7, previousValue) # codebook_zero_value
        for j in range(CbZeroOffset-1, -1, -1):
            self.ue(codebook_egk, previousValue - codebook[j] - 1 ) # codebook_delta_left
            previousValue = codebook[j]
        previousValue = codebook[CbZeroOffset]
        for j in range(CbZeroOffset+1, codebook_size):
            self.ue(codebook_egk, codebook[j] - previousValue - 1 ) # codebook_delta_right
            previousValue = codebook[j]

    def cbZeroOffset( self, codebook_size, CbZeroOffset ):
        codebook_center_offset = CbZeroOffset - (codebook_size >> 1)
        self.ie(2, codebook_center_offset)        

    def entry_point_list(self, block_rows_minus1, dq_flag, cabac_entry_point_list):
        for j in range(block_rows_minus1):
            ep = int(cabac_entry_point_list[j])
            bit_offset  =  ep >> 11
            value       = (ep >>  3) & 255
            dq_state    =  ep & 7
            self.u(8, value)
            if dq_flag:
                self.u(3, dq_state)
            if j == 0:
                self.ue(11, bit_offset)
            else:
                self.ie(7, bit_offset - (int(cabac_entry_point_list[j-1])>>11))

        
class BitReader():
    def __init__(self, byteList):
        assert isinstance( byteList, bytearray )
        self.__byteList = byteList
        self.__bitPos = -1
        self.__bytePos = -1
    
    def readBit(self):
        if self.__bitPos < 0:
            self.__bytePos += 1
            self.__bitPos = 7
        bit = (self.__byteList[self.__bytePos] >> self.__bitPos) & 1
        self.__bitPos -= 1
        return bit
    
    def getNumBytesTouched(self):
        return self.__bytePos + 1

    def getNumBitsTouched(self):
        return (self.__bytePos + 1) * 8 - self.__bitPos - 1

    def u(self, n):
        val = 0
        for i in range( n ):
            val += val + self.readBit()
        return val

    def ue(self, k):
        x = 0
        bit = 1
        while bit:
            bit = 1 - self.u(1)
            x += bit << k
            k += 1

        k -= 1
        if k > 0:
            x += self.u(k)
        return x

    def i(self, n):
        val = self.u( n )
        return val if val < (1<<(n-1)) else val - (1<<n)

    def ie(self,k):
        x = self.ue(k)
        return ((x+1)>>1) if (x&1) else -(x>>1)

    def byte_alignment(self):
        self.u(1)
        self.__bitPos = -1

    def flt(self, n ):
        assert n == 32
        dec_bytes = np.uint32(0)
        dec_bytes += self.u(8)
        dec_bytes += self.u(8) << 8
        dec_bytes += self.u(8) << 16
        dec_bytes += self.u(8) << 24
        z = np.frombuffer( dec_bytes.tobytes(), dtype='<f4', count=1 )
        return z

    def flt_tensor(self, n, dims ):
        assert self.__bitPos == -1
        assert n == 32
        count = np.prod(dims)
        z = np.frombuffer( self.__byteList[self.__bytePos+1:], dtype=np.float32, count=count ).reshape(dims)
        self.__bytePos += 4 * count 
        return z

    def st(self):
        assert self.__bitPos == -1
        remainingBytes = self.__byteList[self.__bytePos+1:]
        strLen = remainingBytes.find( 0 )
        self.__bytePos += strLen + 1
        return  remainingBytes[:strLen].decode( "utf-8", "strict" )

    def codebook(self, codebook_egk, codebook_size, CbZeroOffset):
        codebook = np.zeros(codebook_size, dtype=np.int32)
        previousValue = self.ie(7) # codebook_zero_value
        codebook[CbZeroOffset] = previousValue
        for j in range(CbZeroOffset-1, -1, -1):
            codebook[j] = previousValue - self.ue(codebook_egk) - 1 # codebook_delta_left
            previousValue = codebook[j]
        previousValue = codebook[CbZeroOffset]
        for j in range(CbZeroOffset+1, codebook_size):
            codebook[j] = self.ue(codebook_egk) + previousValue + 1 # codebook_delta_right
            previousValue = codebook[j]
        return codebook

    def cbZeroOffset( self, codebook_size ):
        codebook_center_offset = self.ie(2)
        CbZeroOffset = (codebook_size >> 1) + codebook_center_offset
        return CbZeroOffset
        
    def entry_point_list(self, block_rows_minus1, dq_flag):
        cabac_entry_point_list = np.zeros(block_rows_minus1, dtype=np.uint64)
        for j in range(block_rows_minus1):
            value       = self.u(8)
            dq_state = 0
            if dq_flag:
                dq_state    = self.u(3)
            if j == 0:
                bit_offset = self.ue(11)
            else:
                bit_offset = (int(cabac_entry_point_list[j-1])>>11) + self.ie(7)
            cabac_entry_point_list[j] = (bit_offset<<11) + (value<<3) + dq_state
        return cabac_entry_point_list

class Coder():
    def __init__(self, coder, seDict ):
        self.__seDict = seDict
        self.__coder = coder
        if isinstance( coder, BitReader ):
            self.__isReader = True
        elif isinstance( coder, BitWriter ):
            self.__isReader = False
        else:
            assert 0
    
    def get(self, se ):
        return self.__seDict[se]
    
    def define_array(self, se, dims, dtype ):
        if self.__isReader:
            assert se not in self.__seDict
            if dtype == str:
                assert len( dims ) == 1
                self.__seDict[se] = [ None for _ in range( dims[0] ) ]
            else:
                self.__seDict[se] = np.zeros( dims, dtype=dtype)

    def extend_second_2darray_dimension(self, se, ndl, dtype): ##extend second dimension of 2d array to new dimension ndl
        if self.__isReader:
            assert se in self.__seDict
            assert dtype != str
            assert len(self.__seDict[se].shape) == 2
            cs = self.__seDict[se].shape
            if ndl > cs[1]:
                self.__seDict[se] = np.append( self.__seDict[se], np.zeros( ( cs[0], ndl-cs[1] ), dtype=dtype ), axis=1 )

    def process(self, se, method, *args):
        se_split = se.split( "[" )
        array_idx = None
        array_idx2 = None
        if len( se_split ) > 1:
            se = se_split[0]
            array_idx = int( se_split[1].split("]")[0] )
        if len( se_split ) > 2:
            array_idx2 = int( se_split[2].split("]")[0] )
        m = getattr( self.__coder, method )
        if self.__isReader:
            if array_idx is not None and array_idx2 is not None:
                self.__seDict[se][array_idx][array_idx2] = m( *args )
            elif array_idx is not None:
                self.__seDict[se][array_idx] = m( *args )
            else:
                self.__seDict[se] = m( *args )
        else:
            if array_idx is not None and array_idx2 is not None:
                m( *args, self.__seDict[se][array_idx][array_idx2] )
            elif array_idx is not None:
                m( *args, self.__seDict[se][array_idx] )
            else:
                m( *args, self.__seDict[se] )
    
    def nnr_unit(self, numBytesInNNRUnit ):
        self.nnr_unit_size()
        yield from self.nnr_unit_header()
        self.nnr_unit_payload()

    def nnr_unit_size(self):
        self.process( "nnr_unit_size_flag", "u", 1 )
        self.process( "nnr_unit_size",      "u", 15 + 16 * self.get("nnr_unit_size_flag") )
        
    def nnr_unit_header(self):
        self.process( "nnr_unit_type",                "u", 6 )
        yield
        self.process( "independently_decodable_flag"     , "u", 1 )
        self.process( "partial_data_counter_present_flag", "u", 1)
        if self.get( "partial_data_counter_present_flag" ) == 1:
            self.process( "partial_data_counter", "u", 8 )
        if self.get( "nnr_unit_type" ) == NnrUnitType.NNR_MPS:
            self.nnr_model_parameter_set_unit_header()
        if self.get( "nnr_unit_type" ) == NnrUnitType.NNR_LPS:
            self.nnr_layer_parameter_set_unit_header()
        if self.get( "nnr_unit_type" ) == NnrUnitType.NNR_TPL:
            self.nnr_topology_unit_header()
        if self.get( "nnr_unit_type" ) == NnrUnitType.NNR_QNT:
            self.nnr_quanization_unit_header()
        if self.get( "nnr_unit_type" ) == NnrUnitType.NNR_NDU:
            yield from self.nnr_compressed_data_unit_header()
        if self.get( "nnr_unit_type" ) == NnrUnitType.NNR_STR:
            self.nnr_start_unit_header()
        if self.get( "nnr_unit_type" ) == NnrUnitType.NNR_AGG:
            self.nnr_aggregate_unit_header()

    def nnr_start_unit_header(self):
        self.process( "general_profile_idc", "u", 8 ) ##Not yet specified!

    def nnr_model_parameter_set_unit_header(self):
        ## nnr_reserved_zero_0bit, "u", 0
        pass
    
    def nnr_layer_parameter_set_unit_header(self): ##tbd: implement lps!
        self.process( "lps_self_contained_flag", "u", 1 )
        self.process( "nnr_reserved_zero_7bits", "u", 7 )

    def nnr_topology_unit_header(self):
        self.process( "topology_storage_format", "u", 8 )
        self.process( "topology_compression_format", "u", 8 )
    
    def nnr_quanization_unit_header(self):
        self.process( "quantization_storage_format", "u", 8 )
        self.process( "quantization_compression_format", "u", 8 )

    def codebook(self, cb_suffix):
        self.process( "codebook_egk__"+cb_suffix, "u", 4 )
        self.process( "codebook_size__"+cb_suffix, "ue", 2 )
        self.process( "CbZeroOffset__"+cb_suffix, "cbZeroOffset", self.get( "codebook_size__"+cb_suffix ) )
        self.process( "codebook__"+cb_suffix, "codebook", self.get( "codebook_egk__"+cb_suffix ), self.get( "codebook_size__"+cb_suffix ), self.get( "CbZeroOffset__"+cb_suffix ) )
        
    def nnr_compressed_data_unit_header(self):
        self.process( "nnr_compressed_data_unit_payload_type",       "u", 5 )
        self.process( "nnr_multiple_topology_elements_present_flag", "u", 1 )
        self.process( "nnr_decompressed_data_format_present_flag",   "u", 1 )
        self.process( "input_parameters_present_flag",               "u", 1 )
        
        if self.get("nnr_multiple_topology_elements_present_flag") ==  1:
            self.topology_elements_ids_list(self.get( "mps_topology_indexed_reference_flag" ))
        else:
            if self.get( "mps_topology_indexed_reference_flag" ) == 0: 
                self.process( "topology_elem_id", "st" )
            else:
                self.process( "topology_elem_id_index", "ue", 7 ) 

        if self.get( "general_profile_idc" ) == 1:
            self.process( "node_id_present_flag", "u", 1 )
            if self.get("node_id_present_flag") ==  1:
                self.process( "device_id", "ue", 1 )
                self.process( "parameter_id", "ue", 5 )
                self.process( "put_node_depth", "ue", 4 )

            if self.get( "mps_parent_signalling_enabled_flag" ) == 1:
                self.process( "parent_node_id_present_flag", "u", 1 )
                if self.get("parent_node_id_present_flag") ==  1:
                    self.process( "parent_node_id_type", "u", 2 )
                    self.process( "temporal_context_modeling_flag", "u", 1 )
                    if self.get("parent_node_id_type") == ParentNodeIdType.ICNN_NDU_ID:
                        self.process( "parent_device_id", "ue", 1 )
                        if not self.get("node_id_present_flag") ==  1:
                            self.process( "parameter_id", "ue", 5 )
                            self.process( "put_node_depth", "ue", 4 )
                    elif self.get("parent_node_id_type") == ParentNodeIdType.ICNN_NDU_PL_SHA256:
                        self.process( "parent_node_payload_sha256", "u", 256 )
                    elif self.get("parent_node_id_type") == ParentNodeIdType.ICNN_NDU_PL_SHA512:
                        self.process( "parent_node_payload_sha512", "u", 512 )
                    else:
                        assert 0, "Unknown parent_node_id_type."
        if (
            (self.get( "nnr_compressed_data_unit_payload_type" ) == CompressedDataUnitPayloadType.NNR_PT_FLOAT) or
            (self.get( "nnr_compressed_data_unit_payload_type" ) == CompressedDataUnitPayloadType.NNR_PT_BLOCK)
        ):
             self.process("codebook_present_flag", "u", 1)
             if self.get( "codebook_present_flag" ):
                 self.codebook("")
        
        if (
            (self.get("nnr_compressed_data_unit_payload_type") == CompressedDataUnitPayloadType.NNR_PT_INT) or
            (self.get("nnr_compressed_data_unit_payload_type") == CompressedDataUnitPayloadType.NNR_PT_FLOAT) or
            (self.get("nnr_compressed_data_unit_payload_type") == CompressedDataUnitPayloadType.NNR_PT_BLOCK)
        ):
            self.process("dq_flag", "u", 1)    

        if self.get("nnr_decompressed_data_format_present_flag") == 1:
            self.process( "nnr_decompressed_data_format", "u", 7 )

        if self.get("input_parameters_present_flag") == 1:
            self.process( "tensor_dimensions_flag", "u", 1 )
            self.process( "cabac_unary_length_flag", "u", 1 )
            self.process( "compressed_parameter_types", "u", 4 )
            if self.get( "compressed_parameter_types" ) & BlockParameterTypes.NNR_CPT_DC != 0:
                self.process( "decomposition_rank", "ue", 3 )
                self.process( "g_number_of_rows", "ue", 3 )
            if self.get( "tensor_dimensions_flag" ) == 1:
                self.tensor_dimensions_list()
            if self.get("nnr_compressed_data_unit_payload_type") != CompressedDataUnitPayloadType.NNR_PT_BLOCK:
                if self.get("nnr_multiple_topology_elements_present_flag") == 1:
                    self.topology_tensor_dimension_mapping() ##TODO: To be implemented
            if self.get( "cabac_unary_length_flag" ) == 1:
                self.process( "cabac_unary_length_minus1", "u", 8 )
        yield # pause execution so that tensor dimensions can be properly set based on compressed_parameter_types
        
        if (
            ( self.get("nnr_compressed_data_unit_payload_type") == CompressedDataUnitPayloadType.NNR_PT_BLOCK) and
            ((self.get("compressed_parameter_types") & BlockParameterTypes.NNR_CPT_DC) != 0) and
            ( self.get("codebook_present_flag") )
        ):
                self.codebook("dc")

        if self.get("count_tensor_dimensions") > 1: ##TODO TDR: process first_tensor_dimension_shift here!
                if(self.get("general_profile_idc") == 1):
                    self.process("first_tensor_dimension_shift", "ue", 1)
                self.process( "scan_order", "u", 4 )
                if( self.get("scan_order") > 0 ):
                    tensorDimensions   = self.get( "tensor_dimensions" )
                    blockDim           = 4 << self.get("scan_order")
                    
                    if self.get( "compressed_parameter_types" ) & BlockParameterTypes.NNR_CPT_DC != 0:
                        hNumberOfColumns  = np.int32(np.prod( tensorDimensions )/self.get("g_number_of_rows"))
                        tensorDimensionsG = [self.get("g_number_of_rows"), self.get("decomposition_rank")] 
                        tensorDimensionsH = [self.get("decomposition_rank"), hNumberOfColumns]

                    if (
                        (self.get("nnr_compressed_data_unit_payload_type") != CompressedDataUnitPayloadType.NNR_PT_BLOCK) and
                        (self.get( "compressed_parameter_types" ) & BlockParameterTypes.NNR_CPT_DC != 0)
                    ):
                        if self.get("_decomposed_tensor_type") == "G":
                            tensorDimensions = tensorDimensionsG
                        else:
                            tensorDimensions = tensorDimensionsH

                    numBlockRowsMinus1 = ((tensorDimensions[0]+blockDim-1) >> (2+self.get("scan_order"))) - 1

                    if (
                        (self.get("nnr_compressed_data_unit_payload_type") == CompressedDataUnitPayloadType.NNR_PT_BLOCK) and
                        (self.get( "compressed_parameter_types" ) & BlockParameterTypes.NNR_CPT_DC != 0)
                    ):
                        numBlockRowsMinus1  = ((tensorDimensionsG[0]+blockDim-1) >> (2+self.get("scan_order"))) - 1
                        numBlockRowsMinus1 += ((tensorDimensionsH[0]+blockDim-1) >> (2+self.get("scan_order"))) - 1

                    self.process("cabac_entry_point_list", "entry_point_list", numBlockRowsMinus1, self.get("dq_flag"))            

        self.__coder.byte_alignment()
        
    def tensor_dimensions_list(self):
        self.process( "count_tensor_dimensions", "ue", 1 )
        self.define_array( "tensor_dimensions", [self.get( "count_tensor_dimensions" )], np.uint32 )
        for j in range( self.get( "count_tensor_dimensions" ) ):
            self.process( "tensor_dimensions[%d]" % j, "ue", 7 )

    def topology_elements_ids_list(self, topologyIndexedFlag):
        self.process( "count_topology_elements_minus2", "ue", 7 )
        if topologyIndexedFlag == 0:
            self.__coder.byte_alignment()
            self.define_array( "topology_elem_id_list", [self.get( "count_topology_elements_minus2" ) + 2], str )
        else:
            self.define_array( "topology_elem_id_index_list", [self.get( "count_topology_elements_minus2" ) + 2], np.uint32 )
        for j in range( self.get( "count_topology_elements_minus2" ) + 2 ):
            if topologyIndexedFlag == 0:
                self.process( "topology_elem_id_list[%d]" % j, "st" )
            else:
               self.process( "topology_elem_id_index_list[%d]" % j, "ue", 7 )
        if topologyIndexedFlag == 1:
            self.__coder.byte_alignment()

    def topology_tensor_dimension_mapping(self):
        raise NotImplementedError("topology_tensor_dimension_mapping not yet implemented!")

    def nnr_aggregate_unit_header(self):
        raise NotImplementedError("nnr_aggregate_unit_header not yet implemented!")
    
    def nnr_unit_payload(self):
        if self.get( "nnr_unit_type" ) == NnrUnitType.NNR_MPS:
            self.nnr_model_parameter_set_unit_payload()
        if self.get( "nnr_unit_type" ) == NnrUnitType.NNR_LPS:
            self.nnr_layer_parameter_set_unit_payload()
        if self.get( "nnr_unit_type" ) == NnrUnitType.NNR_TPL:
            self.nnr_topology_unit_payload()
        if self.get( "nnr_unit_type" ) == NnrUnitType.NNR_QNT:
            self.nnr_quanization_unit_payload()
        if self.get( "nnr_unit_type" ) == NnrUnitType.NNR_NDU:
            self.nnr_compressed_data_unit_payload()
        if self.get( "nnr_unit_type" ) == NnrUnitType.NNR_STR:
            self.nnr_start_unit_payload()
        if self.get( "nnr_unit_type" ) == NnrUnitType.NNR_AGG:
            self.nnr_aggregate_unit_payload()

    def nnr_start_unit_payload(self):
        #nnr_reserved_zero_0bit, "u", 0
        pass

    def nnr_model_parameter_set_unit_payload(self):
        self.process( "topology_carriage_flag",    "u",  1 )
        self.process( "mps_sparsification_flag",   "u",  1 )
        self.process( "mps_pruning_flag",          "u",  1 )
        self.process( "mps_unification_flag",      "u",  1 )
        self.process( "mps_decomposition_performance_map_flag", "u",  1 )
        self.process( "mps_quantization_method_flags", "u",  3 )
        self.process( "mps_topology_indexed_reference_flag", "u", 1 )
        if self.get( "general_profile_idc" ) == 1:
            self.process( "base_model_id_present_flag"              , "u", 1 )
            self.process( "validation_set_performance_present_flag" , "u", 1 )
            self.process( "metric_type_performance_map_valid_flag"  , "u", 1 )
            self.process( "mps_parent_signalling_enabled_flag"      , "u", 1 )
            if self.get( "mps_parent_signalling_enabled_flag" ) == 1:
                self.process( "nnr_pre_flag" , "u", 1 )
            else:
                self.process( "nnr_reserved_zero_1bit", "u", 1 )
            self.process( "nnr_reserved_zero_2bits" , "u", 2 )
            if self.get( "base_model_id_present_flag" ) == 1:
                self.process( "base_model_id", "st" )
            if (
                (self.get( "validation_set_performance_present_flag" ) == 1) or
                (self.get( "metric_type_performance_map_valid_flag" )  == 1)
            ):
                self.process( "performance_metric_type", "st" )
        else:
            self.process( "nnr_reserved_zero_7bits", "u", 7 )
        if (
            (self.get( "mps_quantization_method_flags" ) & QuantizationMethodFlags.NNR_QSU != 0) or
            (self.get( "mps_quantization_method_flags" ) & QuantizationMethodFlags.NNR_QCB != 0)
        ):
            self.process( "mps_qp_density",             "u",  3 )
            self.process( "mps_quantization_parameter", "i", 13 )
        if self.get( "mps_sparsification_flag" ) == 1:
            self.sparsification_performance_map()
        if self.get( "mps_pruning_flag" ) == 1:
            self.pruning_performance_map()
        if self.get( "mps_unification_flag" ) == 1:
            self.unification_performance_map()
        if self.get( "mps_decomposition_performance_map_flag" ) == 1:
            self.decomposition_performance_map()
        if (
            (self.get( "general_profile_idc" )                     == 1) and
            (self.get( "validation_set_performance_present_flag" ) == 1)
        ):
            self.process("validation_set_performance", "flt", 32)
        self.__coder.byte_alignment()

    def sparsification_performance_map(self):
        self.process( "spm_count_thresholds", "u", 8 )
        spm_count_threshold_minus1 = self.get("spm_count_thresholds") -1
        self.define_array( "sparsification_threshold", [spm_count_threshold_minus1]    , np.float32 )
        self.define_array( "non_zero_ratio"          , [spm_count_threshold_minus1]    , np.float32 )
        self.define_array( "spm_nn_accuracy"         , [spm_count_threshold_minus1]    , np.float32 )
        self.define_array( "spm_count_classes"       , [spm_count_threshold_minus1]    , np.uint32  )
        self.define_array( "spm_class_bitmask"       , [spm_count_threshold_minus1]    , np.uint32  )
        self.define_array( "spm_nn_class_accuracy"   , [spm_count_threshold_minus1, 1] , np.float32 )
        for i in range( spm_count_threshold_minus1 ):
            self.process( "sparsification_threshold[%d]" % i, "flt", 32 )
            self.process( "non_zero_ratio[%d]" % i          , "flt", 32 )
            self.process( "spm_nn_accuracy[%d]" % i         , "flt", 32 )
            self.process( "spm_count_classes[%d]" % i       , "u"  ,  8 )
            self.process( "spm_class_bitmask[%d]" % i       , "ue" ,  7 )
            self.extend_second_2darray_dimension( "spm_nn_class_accuracy", np.max(self.get("spm_count_classes")), np.float32)
            for j in range( self.get("spm_count_classes")[i] ):
                self.process("spm_nn_class_accuracy[%d][%d]" % (i,j), "flt", 32)

    def pruning_performance_map(self):
        self.process( "ppm_count_pruning_ratios", "u", 8 )
        ppm_count_pruning_ratios_minus1 = self.get("ppm_count_pruning_ratios") -1
        self.define_array( "pruning_ratio"        , [ppm_count_pruning_ratios_minus1]    , np.float32 )
        self.define_array( "ppm_nn_accuracy"      , [ppm_count_pruning_ratios_minus1]    , np.float32 )
        self.define_array( "ppm_count_classes"    , [ppm_count_pruning_ratios_minus1]    , np.uint32 )
        self.define_array( "ppm_class_bitmask"    , [ppm_count_pruning_ratios_minus1]    , np.uint32 )
        self.define_array( "ppm_nn_class_accuracy", [ppm_count_pruning_ratios_minus1, 1] , np.float32 )
        for i in range( ppm_count_pruning_ratios_minus1 ):
            self.process("pruning_ratio[%d]" % i    , "flt", 32)
            self.process("ppm_nn_accuracy[%d]" % i  , "flt", 32)
            self.process("ppm_count_classes[%d]" % i, "u"  ,  8)
            self.process("ppm_class_bitmask[%d]" % i, "ue" ,  7)
            self.extend_second_2darray_dimension( "ppm_nn_class_accuracy", np.max(self.get("ppm_count_classes")), np.float32)
            for j in range( self.get("ppm_count_classes")[i] ):
                self.process("ppm_nn_class_accuracy[%d][%d]" % (i,j), "flt", 32)

    def unification_performance_map(self):
        self.process("upm_count_thresholds", "u", 8)
        upm_count_thresholds_minus1 = self.get("upm_count_thresholds") -1
        self.define_array( "count_reshaped_tensor_dimension" , [upm_count_thresholds_minus1]    , np.uint32  )
        self.define_array( "reshaped_tensor_dimensions"      , [upm_count_thresholds_minus1, 1] , np.uint32  )
        self.define_array( "count_super_block_dimension"     , [upm_count_thresholds_minus1]    , np.uint32  )
        self.define_array( "super_block_dimensions"          , [upm_count_thresholds_minus1, 1] , np.uint32  )
        self.define_array( "count_block_dimension"           , [upm_count_thresholds_minus1]    , np.uint32  )
        self.define_array( "block_dimensions"                , [upm_count_thresholds_minus1, 1] , np.uint32  )
        self.define_array( "unification_threshold"           , [upm_count_thresholds_minus1]    , np.float32 )
        self.define_array( "upm_nn_accuracy"                 , [upm_count_thresholds_minus1]    , np.float32 )
        self.define_array( "upm_count_classes"               , [upm_count_thresholds_minus1]    , np.uint32  )
        self.define_array( "upm_class_bitmask"               , [upm_count_thresholds_minus1]    , np.uint32  )
        self.define_array( "upm_nn_class_accuracy"           , [upm_count_thresholds_minus1, 1] , np.float32 )
        for i in range( upm_count_thresholds_minus1 ):
            self.process( "count_reshaped_tensor_dimension[%d]" % i, "ue", 1)
            self.extend_second_2darray_dimension( "reshaped_tensor_dimensions", np.max(self.get("count_reshaped_tensor_dimension")), np.uint32)
            for j in range( self.get( "count_reshaped_tensor_dimension" )[i] - 1 ):
                self.process( "reshaped_tensor_dimensions[%d][%d]" % (i,j), "ue", 7 )
            self.__coder.byte_alignment()
            self.process( "count_super_block_dimension[%d]" % i, "u", 8)
            self.extend_second_2darray_dimension( "super_block_dimensions", np.max(self.get("count_super_block_dimension")), np.uint32)
            for j in range( self.get( "count_super_block_dimension" )[i] - 1 ):
                self.process( "super_block_dimensions[%d][%d]" % (i,j), "u", 8 )
            self.process( "count_block_dimension[%d]" % i      , "u", 8)
            self.extend_second_2darray_dimension( "block_dimensions", np.max(self.get("count_block_dimension")), np.uint32)
            for j in range( self.get( "count_block_dimension" )[i] - 1 ):
                self.process( "block_dimensions[%d][%d]" % (i,j)      , "u", 8 )
            self.process( "unification_threshold[%d]" % i , "flt", 32 )
            self.process( "upm_nn_accuracy[%d]" % i       , "flt", 32 )
            self.process( "upm_count_classes[%d]" % i     , "u"  ,  8 )
            self.process( "upm_class_bitmask[%d]" % i     , "ue" ,  7 )
            self.extend_second_2darray_dimension( "upm_nn_class_accuracy", np.max(self.get("upm_count_classes")), np.float32)
            for j in range( self.get( "upm_class_bitmask" )[i] ):
                self.process( "upm_nn_class_accuracy[%d][%d]" % (i,j), "flt", 32 )

    def decomposition_performance_map(self):
        self.process("dpm_count_thresholds", "u", 8)
        dpm_count_thresholds_minus1 = self.get("dpm_count_thresholds") - 1
        self.define_array( "mse_threshold"         , [dpm_count_thresholds_minus1]    , np.float32  )
        self.define_array( "dpm_nn_accuracy"       , [dpm_count_thresholds_minus1]    , np.float32  )
        self.define_array( "nn_reduction_ratio"    , [dpm_count_thresholds_minus1]    , np.float32  )
        self.define_array( "dpm_count_classes"     , [dpm_count_thresholds_minus1]    , np.uint32  )
        self.define_array( "dpm_nn_class_accuracy" , [dpm_count_thresholds_minus1, 1] , np.float32  )
        for i in range( dpm_count_thresholds_minus1 ):
            self.process( "mse_threshold[%d]" % i      , "flt", 32 )
            self.process( "dpm_nn_accuracy[%d]" % i    , "flt", 32 )
            self.process( "nn_reduction_ratio[%d]" % i , "flt", 32 )
            self.process( "dpm_count_classes[%d]" % i  , "u"  , 16 )
            self.extend_second_2darray_dimension( "dpm_nn_class_accuracy", np.max(self.get("dpm_count_classes")), np.float32)
            for j in range( self.get( "dpm_count_classes" )[i] ):
                self.process( "dpm_nn_class_accuracy[%d][%d]" % (i,j), "flt", 32 )
    
    def nnr_layer_parameter_set_unit_payload(self):
        self.process( "nnr_reserved_zero_1_bits"      , "u", 1 )
        self.process( "lps_sparsification_flag"       , "u", 1 )
        self.process( "lps_pruning_flag"              , "u", 1 )
        self.process( "lps_unification_flag"          , "u", 1 )
        self.process( "lps_quantization_method_flags" , "u", 3 )
        self.process( "nnr_reserved_zero_1bit"        , "u", 1 )
        if(
            (self.get( "lps_quantization_method_flags" ) & QuantizationMethodFlags.NNR_QCB != 0) or
            (self.get( "lps_quantization_method_flags" ) & QuantizationMethodFlags.NNR_QSU != 0)
        ):
            self.process( "lps_qp_density"            , "u", 3 )
            self.process( "lps_quantization_parameter", "i", 13 )
        if self.get("lps_sparsification_flag") == 1:
            self.sparsification_performance_map()
        if self.get("lps_pruning_flag")        == 1:
            self.pruning_performance_map()
        if self.get("lps_unification_flag")    == 1:
            self.unification_performance_map()
        self.__coder.byte_alignment()

    def nnr_topology_unit_payload(self):
        if self.get( "topology_storage_format" ) == nnr_model.TopologyStorageFormat.NNR_TPL_PRUN:
            self.nnr_pruning_topology_container()
        elif self.get( "topology_storage_format" ) == nnr_model.TopologyStorageFormat.NNR_TPL_REFLIST:
            self.topology_elements_ids_list(0)
        else:
            self.process( "topology_data", "st" )##tbd: implement bs(v) 
    
    def nnr_pruning_topology_container(self):
        raise NotImplementedError("Pruning Topology Container not yet implemented!")

    def nnr_quanization_unit_payload(self):
        self.process( "quantization_data", "st" )

    def nnr_compressed_data_unit_payload(self):
        if self.get( "nnr_compressed_data_unit_payload_type" ) == CompressedDataUnitPayloadType.NNR_PT_RAW_FLOAT:
            self.process( "raw_float32_parameter", "flt_tensor", 32, self.get( "tensor_dimensions" ) )

    def nnr_aggregate_unit_payload(self):
        pass
    
def encode_nnr_unit_with_size_dummy(nnr_unit):
    bs = bytearray()
    w = BitWriter( bs )
    w.u( 32, 0 ) # reserve 4 bytes for nnr_unit_size()
    hls_enc = Coder( w, nnr_unit )
    for _ in hls_enc.nnr_unit_header(): pass
    hls_enc.nnr_unit_payload()
    return bs


def decode_nnr_unit(bitstream, nnr_unit):
    reader = BitReader(bitstream)
    yield from decode_nnr_unit_size_and_header(reader, nnr_unit)
    decode_nnr_unit_payload(reader, nnr_unit)
    yield reader.getNumBytesTouched()

    
def decode_nnr_unit_size_and_header(bit_reader, nnr_unit):
    hls_dec = Coder( bit_reader, nnr_unit )
    hls_dec.nnr_unit_size()
    yield from hls_dec.nnr_unit_header()
    yield

def decode_nnr_unit_payload( bit_reader, nnr_unit_size_and_header ):
    hls_dec = Coder( bit_reader, nnr_unit_size_and_header )
    hls_dec.nnr_unit_payload()


def update_nnr_unit_size( bs ):
    unit_size = {}
    unit_size["nnr_unit_size_flag"] = 1
    if len( bs ) - 2 < 32768:
        # short nnr_unit_size
        bs = bs[2:]
        unit_size["nnr_unit_size_flag"] = 0

    unit_size["nnr_unit_size"] = len( bs )
    w = BitWriter( bs )
    hls_enc = Coder( w, unit_size )
    hls_enc.nnr_unit_size()
    return bs, unit_size



'''
The copyright in this software is being made available under the Clear BSD
License, included below. No patent rights, trademark rights and/or
other Intellectual Property Rights other than the copyrights concerning
the Software are granted under this license.

The Clear BSD License

Copyright (c) 2019-2025, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The NNCodec Authors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted (subject to the limitations in the disclaimer below) provided that
the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

     * Neither the name of the copyright holder nor the names of its
     contributors may be used to endorse or promote products derived from this
     software without specific prior written permission.

NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
'''

from nncodec.nnc_core import hls
import numpy as np
from nncodec.nnc_core import nnr_model

def compile_start_unit(profile):
    ndu_start = {}
    ndu_start["nnr_unit_type"] = hls.NnrUnitType.NNR_STR
    ndu_start["partial_data_counter_present_flag"] = 0
    ndu_start["partial_data_counter"] = 0
    ndu_start["independently_decodable_flag"] = 1

    # nnr_start_unit_header syntax elements
    ndu_start["general_profile_idc"] = profile

    return ndu_start

def compile_mps(approx_data, topology_present, general_profile_idc=0, mps_parent_signalling_enabled_flag=0, pre_signalling=0):
    mps = {}

    # nnr_unit_header syntax elements
    mps["nnr_unit_type"] = hls.NnrUnitType.NNR_MPS
    mps["partial_data_counter_present_flag"] = 0
    mps["partial_data_counter"] = 0
    mps["independently_decodable_flag"] = 1
    
    # model_parameter_set_payload syntax elements
    mps["mps_parent_signalling_enabled_flag"] = mps_parent_signalling_enabled_flag
    mps["topology_carriage_flag"] = topology_present
    mps["mps_sparsification_flag"] = 0
    mps["mps_pruning_flag"] = 0
    mps["mps_unification_flag"] = 0
    mps["mps_decomposition_performance_map_flag"] = 0
    if "qp_density" in approx_data:
        mps["mps_quantization_method_flags"] = hls.QuantizationMethodFlags.NNR_QSU
        mps["mps_qp_density"] = approx_data["qp_density"]
        mps["mps_quantization_parameter"] = 0
    else:
        mps["mps_quantization_method_flags"] = 0

    mps["nnr_pre_flag"] = pre_signalling
    mps["base_model_id_present_flag"] = 0

    mps["mps_topology_indexed_reference_flag"] = 0
    mps["nnr_reserved_zero_7bits"] = 0
    mps["nnr_reserved_zero_1bit"] = 0
    mps["nnr_reserved_zero_2bits"] = 0

    mps["general_profile_idc"] = general_profile_idc

    mps["validation_set_performance_present_flag"] = 0
    mps["metric_type_performance_map_valid_flag"] = 0

    return mps


def compile_ndu_oob(tensor_dims = None, cabac_unary_length_minus1 = None, compressed_parameter_types = None, decomposition_parameter_dict = None ): ##TODO HAASE: OOB several things changed here
    ndu_oob = {}
    ndu_oob["input_parameters_present_flag"] = 1 if not all( [tensor_dims, cabac_unary_length_minus1, compressed_parameter_types, decomposition_parameter_dict and compressed_parameter_types & hls.BlockParameterTypes.NNR_CPT_DC != 0] ) else 0
    if tensor_dims is not None:
        ndu_oob["tensor_dimensions_flag"] = 0
        ndu_oob["tensor_dimensions"] = tensor_dims
        ndu_oob["count_tensor_dimensions"] = len(tensor_dims)
    else:
        ndu_oob["tensor_dimensions_flag"] = 1
    if cabac_unary_length_minus1 is not None:
        ndu_oob["cabac_unary_length_flag"] = 0
        ndu_oob["cabac_unary_length_minus1"] = cabac_unary_length_minus1
    else:
        ndu_oob["cabac_unary_length_flag"] = 1
    if compressed_parameter_types is not None:
        ndu_oob["compressed_parameter_types"] = compressed_parameter_types
    if decomposition_parameter_dict is not None and compressed_parameter_types & hls.BlockParameterTypes.NNR_CPT_DC != 0:
        ndu_oob["decomposition_rank"] = decomposition_parameter_dict["decomposition_rank"]
        ndu_oob["g_number_of_rows"] = decomposition_parameter_dict["g_number_of_rows"]
    return ndu_oob


def compile_ndu_oob_from_dict(oob_dict, ndu, model_info):
    ndu_oob = {}
    if oob_dict == None or len( oob_dict ) == 0:
        return ndu_oob

    if ndu["nnr_multiple_topology_elements_present_flag"] == 1:
        if not ndu["mps_topology_indexed_reference_flag"]:
            param = ndu["topology_elem_id_list"][0] ##All the parameters within a block share the same information, so it's ok to take the first one!
        else:
            topology_elem_id_list_index = ndu["topology_elem_id_index_list"][0]
            param = model_info["topology_elem_id_list"][topology_elem_id_list_index]
    else:
        if not ndu["mps_topology_indexed_reference_flag"]:
            param = ndu["topology_elem_id"]
        else:
            topology_elem_id_index = ndu["topology_elem_id_index"]
            param = model_info["topology_elem_id_list"][topology_elem_id_index]

    if "tensor_dimensions" in oob_dict[param]:
        ndu_oob["tensor_dimensions_flag"] = 0
        ndu_oob["tensor_dimensions"] = oob_dict[param]["tensor_dimensions"]
        ndu_oob["count_tensor_dimensions"] = oob_dict[param]["count_tensor_dimensions"]

    if "cabac_unary_length_minus1" in oob_dict[param]:
        ndu_oob["cabac_unary_length_flag"] = 0
        ndu_oob["cabac_unary_length_minus1"] = oob_dict[param]["cabac_unary_length_minus1"]

    if "compressed_parameter_types" in oob_dict[param]:
        ndu_oob["compressed_parameter_types"] = oob_dict[param]["compressed_parameter_types"]

    if "decomposition_rank" in oob_dict[param] and "g_number_of_rows" in oob_dict[param]:
        ndu_oob["decomposition_rank"] = oob_dict[param]["decomposition_rank"]
        ndu_oob["g_number_of_rows"] = oob_dict[param]["g_number_of_rows"]

    return ndu_oob


def compile_ndu(param, approx_data, enc_info, model_info, ndu_oob, is_block, cpt, block_access, base_params, put_node_depths, tensor_dims=None): ##TODO TDR: add first_tensor_dimension_shift somewhere in the ndu
    ndu_header = {}
    ndu_header.update( ndu_oob )

    if ndu_header["input_parameters_present_flag"] == 1:
       if ndu_header["tensor_dimensions_flag"] == 1:
           assert tensor_dims is not None, "tensor_dimensions must be specified!"
           ndu_header["count_tensor_dimensions"] = len(tensor_dims)
           ndu_header["tensor_dimensions"] = tensor_dims
       if ndu_header["cabac_unary_length_flag"] == 1:
           ndu_header["cabac_unary_length_minus1"] = enc_info.get("cabac_unary_length_minus1", 9)
       ndu_header["compressed_parameter_types"] = cpt
       if cpt & hls.BlockParameterTypes.NNR_CPT_DC != 0:
           ndu_header["decomposition_rank"] = approx_data["decomposition_rank"][block_access.block_id]
           ndu_header["g_number_of_rows"]   = approx_data["g_number_of_rows"][block_access.block_id]
    else:
       assert "tensor_dimensions" in ndu_oob, "tensor_dimensions must be set in out-of-band (oob) parameters"
       assert "cabac_unary_length_minus1" in ndu_oob, "cabac_unary_length_minus1 must be set in out-of-band (oob) parameters"
       assert "compressed_parameter_types" in ndu_oob, "compressed_parameter_types must be set in out-of-band (oob) parameters"
       if cpt & hls.BlockParameterTypes.NNR_CPT_DC != 0:
           assert "decomposition_rank" in ndu_oob, "decomposition_rank must be set in out-of-band (oob) parameters"
           assert "g_number_of_rows" in ndu_oob, "g_number_of_rows must be set in out-of-band (oob) parameters"



    ndu_header["general_profile_idc"] = enc_info.get("general_profile_idc", 0)

    ndu_header["first_tensor_dimension_shift"] = 0 ## currently unused

    ndu_header["mps_topology_indexed_reference_flag"] = 0
    ndu_header["mps_parent_signalling_enabled_flag"] = enc_info.get("mps_parent_signalling_enabled_flag", 0) if enc_info.get("general_profile_idc", 0) else 0

    # nnr_unit_header syntax elements
    ndu_header["nnr_unit_type"] = hls.NnrUnitType.NNR_NDU
    ndu_header["partial_data_counter_present_flag"] = 0 ##tbd: set correctly elsewhere
    ndu_header["partial_data_counter"] = 0
    ndu_header["independently_decodable_flag"] = 1

    # compressed_data_unit_header syntax elements
    if is_block:
        ndu_header["nnr_compressed_data_unit_payload_type"] = hls.CompressedDataUnitPayloadType.NNR_PT_BLOCK
        assert block_access != None , "Block access undefined"
        if enc_info.get("general_profile_idc", 0):
            ndu_header["parameter_id"] = model_info["parameter_index"][block_access.w]
        if cpt & hls.BlockParameterTypes.NNR_CPT_DC != 0:
            param   = block_access.dc_g
            param_h = block_access.dc_h
        else:
            param = block_access.w
    elif bool(approx_data["approx_method"]) and param in approx_data["approx_method"] and \
            ((approx_data["approx_method"][param] == "uniform") or
             (approx_data["approx_method"][param] == "codebook")):
        ndu_header["nnr_compressed_data_unit_payload_type"] = hls.CompressedDataUnitPayloadType.NNR_PT_FLOAT
        if enc_info.get("general_profile_idc", 0):
            if param in model_info["parameter_index"]:
                ndu_header["parameter_id"] = model_info["parameter_index"][param]
            # INCTM (V2): if NNR_PT_BLOCK-parameters are coded separately, i.e., is_block_possible() returns False
            elif param == block_access.bi:
                ndu_header["parameter_id"] = model_info["parameter_index"][block_access.bn_beta]
            elif param == block_access.ls:
                ndu_header["parameter_id"] = model_info["parameter_index"][block_access.bn_gamma]
            else:
                assert 0, "Block parameter not implemented."
    elif approx_data["approx_method"][param] == "skip":
        ndu_header["nnr_compressed_data_unit_payload_type"] = hls.CompressedDataUnitPayloadType.NNR_PT_INT
        if enc_info.get("general_profile_idc", 0):
            ndu_header["parameter_id"] = model_info["parameter_index"][param]
    else:
        assert param not in approx_data["approx_method"], "Unsupported approx_method."
        ndu_header["nnr_compressed_data_unit_payload_type"] = hls.CompressedDataUnitPayloadType.NNR_PT_RAW_FLOAT
        ndu_header["raw_float32_parameter"] = approx_data["parameters"][param]
        if enc_info.get("general_profile_idc", 0):
            ndu_header["parameter_id"] = model_info["parameter_index"][param]

    assert param != None 

    if (
        (ndu_header["nnr_compressed_data_unit_payload_type"] == hls.CompressedDataUnitPayloadType.NNR_PT_BLOCK) or
        (ndu_header["nnr_compressed_data_unit_payload_type"] == hls.CompressedDataUnitPayloadType.NNR_PT_FLOAT) or
        (ndu_header["nnr_compressed_data_unit_payload_type"] == hls.CompressedDataUnitPayloadType.NNR_PT_INT)
    ):
        ndu_header["dq_flag"] = approx_data["dq_flag"][param]

    ndu_header["nnr_multiple_topology_elements_present_flag"] = 1 if ndu_header["nnr_compressed_data_unit_payload_type"] == hls.CompressedDataUnitPayloadType.NNR_PT_BLOCK else 0
    ndu_header["nnr_decompressed_data_format_present_flag"] = 0
    ndu_header["input_parameters_present_flag"] = 1
    ndu_header["compressed_parameter_types"] = cpt
    if cpt & hls.BlockParameterTypes.NNR_CPT_DC != 0:
        ndu_header["decomposition_rank"] = approx_data["decomposition_rank"][block_access.block_id]
        ndu_header["g_number_of_rows"]   = approx_data["g_number_of_rows"][block_access.block_id]

    if enc_info["cabac_unary_length_minus1"] !=  ndu_header["cabac_unary_length_minus1"]:
        # move from out-of-band to in-band
        ndu_header["cabac_unary_length_minus1"] = enc_info["cabac_unary_length_minus1"]
        ndu_header["input_parameters_present_flag"] = 1
        ndu_header["cabac_unary_length_flag"] = 1

    if ndu_header["nnr_multiple_topology_elements_present_flag"] == 1:
        tpl_elem_ids = [p for p in block_access.topology_elem_generator(approx_data["compressed_parameter_types"])]
        if ndu_header["mps_topology_indexed_reference_flag"] == 1:
            assert 0, "topology indexed not implemented yet!"
        else:
            ndu_header["count_topology_elements_minus2"] = len(tpl_elem_ids) - 2
            ndu_header["topology_elem_id_list"] = tpl_elem_ids
    else:
        if ndu_header["mps_topology_indexed_reference_flag"] == 1:
            assert 0, "topology indexed not implemented yet!"
        ndu_header["topology_elem_id"] = param #topology_elem_id == param Name

    if approx_data["approx_method"][param] == "codebook":
        ndu_header["codebook_present_flag"] = 1
        ndu_header["codebook_egk__"] = approx_data["codebooks_egk"][param]
        codebook_size = len( approx_data["codebooks"][param] )
        ndu_header["CbZeroOffset__"] = approx_data["codebook_zero_offsets"][param]
        ndu_header["codebook_size__"] = codebook_size
        ndu_header["codebook__"] = approx_data["codebooks"][param]
        if is_block and (cpt & hls.BlockParameterTypes.NNR_CPT_DC != 0):
            assert approx_data["approx_method"][param_h] == "codebook", "Params must have the same approx_method!"
            ndu_header["codebook_egk__dc"] = approx_data["codebooks_egk"][param_h]
            codebook_size = len( approx_data["codebooks"][param_h] )
            ndu_header["CbZeroOffset__dc"] = approx_data["codebook_zero_offsets"][param_h]
            ndu_header["codebook_size__dc"] = codebook_size
            ndu_header["codebook__dc"] = approx_data["codebooks"][param_h]
    else:
        ndu_header["codebook_present_flag"] = 0

    if enc_info.get("general_profile_idc", 0):
        ndu_header["node_id_present_flag"] = enc_info["node_id_present_flag"]
        if enc_info["node_id_present_flag"]:
            ndu_header["device_id"] = enc_info["device_id"]
            if put_node_depths is None:
                ndu_header["put_node_depth"] = 0 # parameter of base model
            elif param not in put_node_depths:
                ndu_header["put_node_depth"] = 1 # first incremental update
            else:
                ndu_header["put_node_depth"] = put_node_depths[param] + 1 # One more than the parent
        ndu_header["parent_node_id_present_flag"] = enc_info["parent_node_id_present_flag"]
        if enc_info["parent_node_id_present_flag"]:
            ndu_header["parent_node_id_type"] = enc_info["parent_node_id_type"]
            ndu_header["temporal_context_modeling_flag"] = 1 if (enc_info["temporal_context_modeling_flag"] and
                                                                    base_params != None and
                                                                    param in base_params["parameters"]
                                                                    ) else 0
            if enc_info["parent_node_id_type"] == hls.ParentNodeIdType.ICNN_NDU_ID:
                ndu_header["parent_device_id"] = enc_info["parent_device_id"]
                assert put_node_depths is not None
                if not enc_info["node_id_present_flag"]:
                    ndu_header["put_node_depth"] = put_node_depths[param] + 1 if param in put_node_depths else 1
            else:
                assert 0, "Not implemented."

    ndu_header["nnr_decompressed_data_format"] = hls.DecompressedDataFormat.TENSOR_FLOAT32

    if len( ndu_header["tensor_dimensions"] ) > 1: ##TODO TDR: probably add first_tensor_dimension_shift here!
        ndu_header["scan_order"] = approx_data["scan_order"][param]

    return ndu_header


def compile_ndu_eps( ndu_header, cabac_entry_point_list ):
    dims = ndu_header["tensor_dimensions"]
    height = dims[0]
    width = np.prod(dims[1:])
    if width > 1 and height > 1:
#    if len( ndu_header["tensor_dimensions"] ) > 1:
        if ndu_header["scan_order"] > 0:
#            ndu_header["cabac_entry_point_list_size"] = len(cabac_entry_point_list)
            ndu_header["cabac_entry_point_list"] = cabac_entry_point_list

    return ndu_header

def compile_tpl(model_info):
    if model_info["topology_storage_format"] == nnr_model.TopologyStorageFormat.NNR_TPL_UNREC: ##ignore topology_data in this case!
        tpl = {
            "topology_data": "",
            "topology_storage_format": nnr_model.TopologyStorageFormat.NNR_TPL_UNREC
        }
    elif model_info["topology_storage_format"] == nnr_model.TopologyStorageFormat.NNR_TPL_PYT: ##topology_data can be ignored. The parameter names are transmitted as topology_elem_id/topology_elem_id_index anyway!
        topology_data = ''
        tpl = {
            "topology_data": topology_data,
            "topology_storage_format": nnr_model.TopologyStorageFormat.NNR_TPL_PYT
        }
    elif model_info["topology_storage_format"] == nnr_model.TopologyStorageFormat.NNR_TPL_TEF: ##topology_data can be ignored. The parameter names are transmitted as topology_elem_id/topology_elem_id_index anyway!
        topology_data = ''
        tpl = {
            "topology_data": topology_data,
            "topology_storage_format": nnr_model.TopologyStorageFormat.NNR_TPL_TEF
        }
    else:
        raise NotImplementedError("The given topology storage format is not implemented.")

    tpl.update({
            "nnr_unit_type": hls.NnrUnitType.NNR_TPL,
            "partial_data_counter_present_flag": 0,
            "partial_data_counter": 0,
            "independently_decodable_flag": 1,
            "mps_topology_indexed_reference_flag": 0,
            "topology_compression_format" : model_info["topology_compression_format"],
    })
    return tpl




'''
The copyright in this software is being made available under the Clear BSD
License, included below. No patent rights, trademark rights and/or 
other Intellectual Property Rights other than the copyrights concerning 
the Software are granted under this license.

The Clear BSD License

Copyright (c) 2019-2025, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The NNCodec Authors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted (subject to the limitations in the disclaimer below) provided that
the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

     * Neither the name of the copyright holder nor the names of its
     contributors may be used to endorse or promote products derived from this
     software without specific prior written permission.

NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
'''

import copy
import enum
from abc import ABC, abstractmethod
from nncodec import nnc_core
import numpy as np
from nncodec.nnc_core import hls, common
import pickle

class TopologyStorageFormat(enum.IntEnum):
    NNR_TPL_UNREC   = 0
    NNR_TPL_NNEF    = 1
    NNR_TPL_ONNX    = 2
    NNR_TPL_PYT     = 3
    NNR_TPL_TEF     = 4
    NNR_TPL_PRUN    = 5
    NNR_TPL_REFLIST = 6
    
class TopologyCompressionFormat(enum.IntEnum):
    NNR_PT_RAW = 0
    NNR_DFL    = 1

W_TYPES = [
    "weight"
]
O_TYPES = [
    "weight.ls",
    "bias",
    "bn.beta",
    "bn.gamma",
    "bn.mean",
    "bn.var",
    "unspecified"
]

class ModelExecute(ABC):
    def eval_model(self,
                   parameters,
                   verbose=False,
                   ):
        
        raise NotImplementedError("Function eval_model not yet implemented. This function is e.g. required for Inference-optimised quantization (IOQ). Either implement eval_model or deactivate IOQ (controlled by parameter ioq)!")


    def test_model(self,
                   parameters,
                   verbose=False,
                   ):
        
        raise NotImplementedError("Function test_model not yet implemented. This function is e.g. required for inference.")
    
    
    def tune_model(self,
                   parameters,
                   param_types,
                   lsa_flag,
                   ft_flag,
                   verbose=False,
                   ):
        
        raise NotImplementedError("Function tune_model not yet implemented. This function is required for fine tuning (FT) and local scaling adaptation (LSA). Either implement the function or deactivate fine tuning (controlled by parameter ft) and local scaling adaptation (controlled by parameter lsa)!")
    
    @abstractmethod
    def has_eval(self):
        return False
    
    @abstractmethod
    def has_test(self):
        return False
    
    @abstractmethod
    def has_tune_ft(self):
        return False
    
    @abstractmethod
    def has_tune_lsa(self):
        return False
    
    
def create_NNC_model_instance_from_file( model_path ):

    try:
        parameter_dict = pickle.load( model_path )
 
        if isinstance(parameter_dict, dict) and all( [isinstance(a, np.ndarray) for a in parameter_dict.values()] ) and (all([a.dtype==np.float32 for a in parameter_dict.values()]) or  all([a.dtype==np.int32 for a in parameter_dict.values()])):
            NNCModel = NNRModel( parameter_dict )
        else:
            raise SystemExit("Parameter dict must be a dict (key-value pairs). The keys shall be stings, specifying the tensor names. The values must be numpy arrays (ndarray) of type float32 or int32!")
    except:
        raise SystemExit("Can't read model: {}".format(model_path))

    return NNCModel, parameter_dict


def save_to_pickled_file( model_dict, model_path ):
    pickle.dump( model_dict, open(model_path, "wb") )

class NNRModel():

    def __init__(self, model_dict=None):
        
        if model_dict and isinstance(model_dict, dict):
            self.init_model_from_dict(model_dict)
        else:
            self.__model_info = None

        self.model = None


    def init_model_from_dict(self, model_dict):
        
        if isinstance(model_dict, dict):
            model_dict = model_dict
        else:
            raise SystemExit("model_dict must be of type dict")

        model_data = {'parameters': {}, 'reduction_method': 'baseline'}
        model_info = {'parameter_type': {}, 'parameter_dimensions': {}, 'parameter_index': {}, 'block_identifier': {}, 'original_size': {}, 'topology_storage_format' : None, 'topology_compression_format' : None}

        type_list_int = ['int8', 'int16', 'int32', 'uint8', 'uint16', 'uint32']
        type_list_1_bytes = ['int8', 'uint8']
        type_list_2_bytes = ['int16', 'uint16', 'float16']
        original_size = 0

        for i, module_name in enumerate(model_dict):
            if model_dict[module_name].dtype in type_list_1_bytes:
                original_size += model_dict[module_name].size
            elif model_dict[module_name].dtype in type_list_2_bytes:
                original_size += model_dict[module_name].size*2
            else:
                original_size += model_dict[module_name].size*4
            model_data['parameters'][module_name] = np.int32(model_dict[module_name]) if model_dict[module_name].dtype in type_list_int else model_dict[module_name]
            mdl_shape = model_data['parameters'][module_name].shape
            model_info['parameter_dimensions'][module_name] = mdl_shape
            if len(mdl_shape) == 0: #scalar
                model_data['parameters'][module_name] = np.array([np.float32(model_data['parameters'][module_name])])
                model_info['parameter_dimensions'][module_name] = np.array([0]).shape
            model_info['parameter_index'][module_name] = i

            dims = len(mdl_shape)

            if dims > 1:
                model_info['parameter_type'][module_name] = 'weight'
            else:
                model_info['parameter_type'][module_name] = 'unspecified'

        model_info['topology_storage_format'] = nnc_core.nnr_model.TopologyStorageFormat.NNR_TPL_UNREC
        model_info['topology_compression_format'] = nnc_core.nnr_model.TopologyCompressionFormat.NNR_PT_RAW

        model_info["original_size"] = original_size

        self.__model_info = model_info

        return model_data["parameters"]
    
    def guess_block_id_and_param_type(self, model_parameters):
        raise SystemExit("Block id and parameter type can not be guessed for generic model class. Try to provide a pytorch model, a tensorflow model or block_id_and_param_type (see description of compress_model)!")
                    
    @property
    def model_info(self):
        return self.__model_info


class NNRParamAccess():
    def __init__(self, model_info, param):
        self.__single_param = (model_info["parameter_type"].get(param), param, model_info["parameter_dimensions"].get(param))

    def param_generator(self, cpt_dict_dummy):
        yield self.__single_param

    @property
    def block_id(self):
        return None

    @property
    def param(self):
        return self.__single_param[1]
    

class NNRBlockAccess():
    def __init__(self, model_info, block_identifier):
        self.__block_identifier = block_identifier
        self.__model_info = model_info
        block_list = [
            x
            for x in model_info["block_identifier"]
            if model_info["block_identifier"][x] == block_identifier
        ] 
        self.__block_dict = { model_info["parameter_type"][x]: x for x in block_list }


    @property
    def block_id(self):
        return self.__block_identifier

    @property
    def w(self):
        for x in ["weight"]:
            if x in self.__block_dict:
                return self.__block_dict[x]

    @property
    def dc_g(self):
        return self.w + "_G"

    @property
    def dc_h(self):
        return self.w + "_H"

    @property
    def ls(self):
        return self.w + "_scaling"
            
    @property
    def bn_beta(self):
        return self.__block_dict.get("bn.beta", None)

    @property
    def bn_gamma(self):
        return self.__block_dict.get("bn.gamma", None)

    @property
    def bn_mean(self):
        return self.__block_dict.get("bn.mean", None)

    @property
    def bn_var(self):
        return self.__block_dict.get("bn.var", None)

    @property
    def bi(self):
        for x in ["bias"]:
            if x in self.__block_dict:
                return self.__block_dict[x]
        for x in ["weight"]:
            if x in self.__block_dict:
                return self.__block_dict[x] + ".bias"

    
    def param_generator(self, compressed_parameter_types_dict):
        compressed_parameter_types = compressed_parameter_types_dict[self.block_id]

        if compressed_parameter_types & hls.BlockParameterTypes.NNR_CPT_LS != 0: 
            yield "weight.ls", self.ls, [self.__model_info["parameter_dimensions"][self.w][0]]
        if compressed_parameter_types & hls.BlockParameterTypes.NNR_CPT_BI != 0:
            yield "bias", self.bi, [self.__model_info["parameter_dimensions"][self.w][0]]
        if compressed_parameter_types & hls.BlockParameterTypes.NNR_CPT_BN != 0:
            yield "bn.beta", self.bn_beta, self.__model_info["parameter_dimensions"][self.bn_beta]
            yield "bn.gamma", self.bn_gamma, self.__model_info["parameter_dimensions"][self.bn_gamma]
            yield "bn.mean", self.bn_mean, self.__model_info["parameter_dimensions"][self.bn_mean]
            yield "bn.var", self.bn_var, self.__model_info["parameter_dimensions"][self.bn_var]
        if compressed_parameter_types & hls.BlockParameterTypes.NNR_CPT_DC != 0:
            yield "weight", self.dc_g, self.__model_info["parameter_dimensions"][self.w]
            yield "weight", self.dc_h, self.__model_info["parameter_dimensions"][self.w]
        else:
            yield "weight", self.w, self.__model_info["parameter_dimensions"][self.w]
            
    def topology_elem_generator(self, compressed_parameter_types_dict):
        compressed_parameter_types = compressed_parameter_types_dict[self.block_id]
        if compressed_parameter_types & hls.BlockParameterTypes.NNR_CPT_DC != 0:
            yield self.dc_g
            yield self.dc_h
        else:
            yield self.w
        if compressed_parameter_types & hls.BlockParameterTypes.NNR_CPT_LS != 0: 
            yield  self.ls
        if compressed_parameter_types & hls.BlockParameterTypes.NNR_CPT_BN != 0:
            yield self.bn_beta
            yield self.bn_gamma
            yield self.bn_mean
            yield self.bn_var
        if compressed_parameter_types & hls.BlockParameterTypes.NNR_CPT_BI != 0:
            yield self.bi
        

class NNRModelAccess():
    def __init__(self, model_info):
        self.__model_info = model_info
        self.__block_list = []
        block_set_check = set( model_info["block_identifier"].values() )
        params_sorted = sorted(model_info["parameter_index"], key=model_info["parameter_index"].get)
        for param in params_sorted:
            if param in model_info["block_identifier"]:
                if model_info["parameter_type"][param] in ["weight"]:
                    self.__block_list.append( (model_info["block_identifier"][param], param) )
                    block_set_check.remove( model_info["block_identifier"][param] )
            else:
                self.__block_list.append( (None, param) )
        assert not block_set_check, "Unresolved block identifiers: {}".format( block_set_check )

    def blocks_and_params(self):
        for block_id, param in self.__block_list:
            if block_id is None:
                yield NNRParamAccess(self.__model_info, param)
            else:
                yield NNRBlockAccess(self.__model_info, block_id)

        

def set_block_id_and_param_type(model_info, block_id_and_param_type):
    assert "block_identifier" in block_id_and_param_type, "block_identifier not available!"
    assert "parameter_type" in block_id_and_param_type, "parameter_type not available!"
    
    model_info["block_identifier"] = {}
    
    block_id_values_list = list(block_id_and_param_type["block_identifier"].values())

    for param, pardIdx in model_info["parameter_index"].items():
        model_info["parameter_index"][param] = pardIdx
        if param in block_id_and_param_type["parameter_type"].keys():
            model_info["parameter_type"][param] = block_id_and_param_type["parameter_type"][param]
        if param in block_id_and_param_type["block_identifier"].keys() and block_id_and_param_type["block_identifier"][param] is not None and block_id_values_list.count(block_id_and_param_type["block_identifier"][param]) > 1 :
            model_info["block_identifier"][param] = block_id_and_param_type["block_identifier"][param]
            
            
def add_lsa_to_block_id_and_param_type( block_id_and_param_type, lsa_params ):
    for key in lsa_params.keys():
        if key not in block_id_and_param_type["block_identifier"]:
            block_id_and_param_type["block_identifier"][key] = block_id_and_param_type["block_identifier"].get(key.strip("_scaling"), None)
            block_id_and_param_type["parameter_type"][key] = "weight.ls"
            

def sanity_check_block_id_and_param_type(block_id_and_param_type, model_parameters=None):
    block_dict = dict()
    sanity_check_success = True
    for param, blkId in block_id_and_param_type["block_identifier"].items():
        if blkId != None:
            parT = block_id_and_param_type["parameter_type"][param]
            parShape = model_parameters[param].shape if model_parameters else None
            if model_parameters and parT != "weight" and len(parShape) != 1:
                sanity_check_success = False
                break
            if blkId not in block_dict.keys():
                block_dict[blkId] = []
            block_dict[blkId].append([param, parT, parShape])
    
    for bId, bList in block_dict.items():
        available_types = ["weight", "weight.ls", "bias", "bn.mean", "bn.var", "bn.gamma", "bn.beta"]
        lastShape = None
        for par, parT, parShape in bList:
            if parT not in available_types and parT != "unspecified":
                sanity_check_success = False
                break
            if parT != "unspecified":
                available_types.remove(parT) 
            if lastShape != None and lastShape[0] != parShape[0]:
                sanity_check_success = False
                break
            lastShape = parShape
        if "weight" in available_types:
            sanity_check_success = False
            break
    
    return sanity_check_success


'''
The copyright in this software is being made available under the Clear BSD
License, included below. No patent rights, trademark rights and/or
other Intellectual Property Rights other than the copyrights concerning
the Software are granted under this license.

The Clear BSD License

Copyright (c) 2019-2025, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The NNCodec Authors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted (subject to the limitations in the disclaimer below) provided that
the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

     * Neither the name of the copyright holder nor the names of its
     contributors may be used to endorse or promote products derived from this
     software without specific prior written permission.

NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
'''
import copy
import os
import torch
import torchvision
import numpy as np
import wandb
from collections import OrderedDict
import flwr as fl
from typing import Dict, List, Optional, Tuple, Union
from flwr.common import (FitRes, Parameters, Scalar, ndarrays_to_parameters, parameters_to_ndarrays)
from flwr.server.client_proxy import ClientProxy
from flwr.server.strategy.aggregate import aggregate

def model_add(state_dict_a, state_dict_b): # full, partial
    state_dict_sum = OrderedDict()
    for param_name in state_dict_a:
        if param_name in state_dict_b:
            state_dict_sum[param_name] = state_dict_a[param_name] + state_dict_b[param_name]
        else:
            state_dict_sum[param_name] = state_dict_a[param_name]
    return state_dict_sum


class NNCFedAvg(fl.server.strategy.FedAvg):

    def __init__(self, min_fit_clients, min_available_clients, evaluate_fn, accept_failures,
                 fit_metrics_aggregation_fn, initial_parameters, model_arch, mdl_info,  args, encode_fn, decode_fn):

        super().__init__() # Call the constructor of the original class
        self.id = 0
        self.min_fit_clients = min_fit_clients
        self.min_available_clients = min_available_clients
        self.evaluate_fn = evaluate_fn
        self.accept_failures = accept_failures
        self.initial_parameters = initial_parameters
        self.fit_metrics_aggregation_fn = fit_metrics_aggregation_fn
        self.accumulated_bs_size = 0
        self.accumulated_uncompressed = 0
        self.current_bs_size = 0
        self.model_arch = model_arch
        self.mdl_info = mdl_info
        self.bytes_mdl_full_prec = sum([v.numel( ) * 4 for v in self.model_arch.state_dict().values() if v.shape != torch.Size([])])
        self.args = args
        self.encode_fn = encode_fn
        self.decode_fn = decode_fn
        self.base_mdl = OrderedDict({k: np.float32(v.cpu().detach().numpy()) for k, v in self.model_arch.state_dict().items()
                                    if v.shape != torch.Size([])})
        if self.args.bnf:
            self.base_mdl = self.fold_bn(self.base_mdl)
        self.expected_keys = self.mdl_info["parameter_index"] if not self.args.bnf else mdl_info["bnf_map"]
        self.internal_states = {}
        self.internal_states = {"approx_param_base": {"parameters": {},
                                                      "put_node_depth": {},
                                                      "device_id": self.id,
                                                      "parameter_id": {},
                                                      },
                                "best_loss": 1e9}

        self.previously_encoded_params = None

        print("SERVER INITIALIZED") ## server only once, clients every time they're called

    def evaluate(
            self, server_round: int, parameters: Parameters
    ) -> Optional[Tuple[float, Dict[str, Scalar]]]:
        """Evaluate model parameters using an evaluation function."""

        if self.evaluate_fn is None:
            # No evaluation function provided
            return None

        if self.args.compress_downstream and len(parameters.tensors) == 1: # NNC-encoded
            parameters_ndarrays = self.decode_fn(bytearray(parameters.tensors[0][parameters.tensors[0].index(b'\n') + 1:]),
                                                 internal_states_path=f"{self.args.results}")
            self.save_internal_states()

            if self.args.compress_differences: ## TODO: replace w/ checking mps_parent_signalling_enabled_flag
                parameters_ndarrays = model_add(self.base_mdl, parameters_ndarrays)
                self.base_mdl = copy.deepcopy(parameters_ndarrays)

            if self.args.bnf:
                parameters_ndarrays = self.unfold_bn(parameters_ndarrays)

            state_dict = OrderedDict({k: torch.Tensor(v) for k, v in parameters_ndarrays.items()})

        else: # not NNC-encoded
            parameters_ndarrays = parameters_to_ndarrays(parameters)
            state_dict = OrderedDict({k: torch.Tensor(v) for k, v in zip(self.mdl_info["parameter_index"], parameters_ndarrays)})

        test_results_dict = self.evaluate_fn(server_round, state_dict)

        if self.args.wandb:
            print("wandb: log accumulated_bs_size and global test performance")
            is_seg_mdl = self.args.model in torchvision.models.list_models(torchvision.models.segmentation)
            wandb_dict = {"accumulated_bs_size": self.accumulated_bs_size,
                          "accumulated_uncompressed_size": self.accumulated_uncompressed,
                          "running_CR": self.accumulated_bs_size / self.accumulated_uncompressed if self.accumulated_uncompressed else 0,
                          "current_bs_size": self.current_bs_size,
                          "current_CR": self.current_bs_size / (self.bytes_mdl_full_prec * self.min_available_clients * 2),
                          }
            wandb_dict.update(test_results_dict)
            wandb.log(wandb_dict)

        mean_loss = torch.mean(torch.tensor(list(test_results_dict.values())))
        if mean_loss <= self.internal_states["best_loss"]:
           self.internal_states["best_loss"] = mean_loss
           print("save new best model")
           torch.save(state_dict, os.path.join(self.args.results, f'best_{self.args.model}_{self.args.job_id}.pt'))

        return mean_loss, test_results_dict


    def aggregate_fit(
            self,
            server_round: int,
            results: List[Tuple[ClientProxy, FitRes]],
            failures: List[Union[Tuple[ClientProxy, FitRes], BaseException]],
    ) -> Tuple[Optional[Parameters], Dict[str, Scalar]]:
        """Aggregate fit results using weighted average."""

        if not results:
            return None, {}
        # Do not aggregate if there are failures and failures are not accepted
        if not self.accept_failures and failures:
            return None, {}

        if os.path.exists(f"{self.args.results}/client_ID{self.id}_internal_states.npz"):
            self.load_internal_states()

        if self.previously_encoded_params:

            _ = self.decode_fn(self.previously_encoded_params,
                                           approx_param_base=self.internal_states["approx_param_base"],
                                           update_base_param=True)

        # Convert results
        if self.args.compress_upstream:
            decompressed_weights = [(self.decode_fn(bytearray(bs[bs.index(b'\n') + 1:]),
                                                    internal_states_path=f"{self.args.results}",
                                                    update_base_param=True), ne)
                                                    for bs, ne in [(fitres.parameters.tensors[0], fitres.num_examples)
                                                        for _, fitres in results]]

            # If any key of the expected keys is missing, it is inserted with a zero valued ndarray
            for tensors, num_samples in decompressed_weights:
                if len(tensors) != len(self.expected_keys):
                    for k in self.expected_keys.keys():
                        tensors.setdefault(k, np.zeros_like(self.base_mdl[k]))

            # sort all client updates in the same order (as in self.expected_keys)
            decompressed_weights = [(dict(sorted(tensors.items(), key=lambda x: list(self.expected_keys.keys()).index(x[0]))), num_samples)
                                    for tensors, num_samples in decompressed_weights]

            weights_results = [([v for _, v in w.items() if v.shape != ()], ne) for w, ne in decompressed_weights]

        else:
            weights_results = [
                (parameters_to_ndarrays(fit_res.parameters), fit_res.num_examples)
                for _, fit_res in results
            ]

        # Aggregate
        parameters_aggregated = aggregate(weights_results)

        if self.args.compress_downstream:
            print("DOWN-STREAM compression:")
            agg_mdl_state_dict = OrderedDict({k: np.float32(v) for k, v in zip(self.expected_keys, parameters_aggregated)})
            
            if self.args.err_accumulation and "residuals" in self.internal_states:
                agg_mdl_state_dict = model_add(agg_mdl_state_dict, self.internal_states["residuals"])

            parameters_aggregated = [self.encode_fn(agg_mdl_state_dict, vars(self.args), approx_param_base=self.internal_states["approx_param_base"],
                                                    device_id=self.id)]
            self.accumulated_bs_size += (self.min_fit_clients * len(parameters_aggregated[0]))
            current_bs_size_server = self.min_fit_clients * len(parameters_aggregated[0])

            self.previously_encoded_params = parameters_aggregated[0]
            
            if self.args.err_accumulation:
                self.update_residual(compressed_update=self.decode_fn(parameters_aggregated[0], approx_param_base=self.internal_states["approx_param_base"]),
                                     uncompressed_update=agg_mdl_state_dict)
        else:
            self.accumulated_bs_size += (self.min_fit_clients * self.bytes_mdl_full_prec)
            current_bs_size_server = self.min_fit_clients * self.bytes_mdl_full_prec

        self.save_internal_states()

        parameters_aggregated = ndarrays_to_parameters(parameters_aggregated)

        self.accumulated_uncompressed += self.bytes_mdl_full_prec * self.min_available_clients * 2 # up and down link

        # Aggregate custom metrics if aggregation fn was provided
        metrics_aggregated = {}
        if self.fit_metrics_aggregation_fn:
            fit_metrics = [(res.num_examples, res.metrics) for _, res in results]
            metrics_aggregated = self.fit_metrics_aggregation_fn(fit_metrics)
            self.accumulated_bs_size += metrics_aggregated["accumulated_bs_sizes"]
            self.current_bs_size = current_bs_size_server + metrics_aggregated["accumulated_bs_sizes"]
            print(f"accumulated_bs_size: {self.accumulated_bs_size}")
        return parameters_aggregated, metrics_aggregated
    
    def update_residual(self, compressed_update, uncompressed_update):
        if not "residuals" in self.internal_states:
            self.internal_states["residuals"] = {}
        for k in uncompressed_update:
            ignored_param = False #k.endswith(".weight_scaling")
            if k in compressed_update and not ignored_param:
                self.internal_states["residuals"][k] = uncompressed_update[k] - compressed_update[k]
            else:
                self.internal_states["residuals"][k] = np.zeros_like(uncompressed_update[k]) \
                                                        if ignored_param else uncompressed_update[k]

    def fold_bn(self, param_dict):
        eps = 1e-3 if self.mdl_info['topology_storage_format'] == 4 else 1e-5

        for blk in self.mdl_info["mdl_blocks"]:
            block_access = self.mdl_info["mdl_blocks"][blk]
            block_id = block_access["block_id"]
            if block_id is None:
                continue

            if "bn_mean" in block_access:
                bn_shape = param_dict[block_access["bn_mean"]].shape

                delta = block_access["bi"]
                if delta not in param_dict:
                    param_dict[delta] = np.zeros(bn_shape, dtype=np.float32)

                alpha = block_access["ls"]
                if alpha not in param_dict:
                    param_dict[alpha] = np.ones(bn_shape, dtype=np.float32)

                g = param_dict[block_access["bn_gamma"]] / np.sqrt(param_dict[block_access["bn_var"]] + eps)
                param_dict[alpha] *= g
                param_dict[delta] = (param_dict[delta] - param_dict[block_access["bn_mean"]]) * g + param_dict[block_access["bn_beta"]]

                del param_dict[block_access["bn_gamma"]], param_dict[block_access["bn_var"]],\
                    param_dict[block_access["bn_mean"]], param_dict[block_access["bn_beta"]]

        return param_dict
    def unfold_bn(self, folded_state_dict):
        mdl_arch = {n: v.detach().cpu().numpy() for n, v in self.model_arch.state_dict().items()}
        unfolded_state_dict = {}

        for blk in self.mdl_info["mdl_blocks"]:
            block_access = self.mdl_info["mdl_blocks"][blk]
            if block_access["block_id"] is None:
                continue

            if "w" in block_access and block_access["w"] in mdl_arch:
                unfolded_state_dict[block_access["w"]] = folded_state_dict[block_access["w"]]
            if "bi" in block_access and block_access["bi"] in mdl_arch:
                unfolded_state_dict[block_access["bi"]] = folded_state_dict[block_access["bi"]]
            if "bn_mean" in block_access and block_access["bn_mean"] in mdl_arch:
                bn_shape = mdl_arch[block_access["bn_mean"]].shape
                unfolded_state_dict[block_access["bn_mean"]] = np.zeros(bn_shape, dtype=np.float32)
                unfolded_state_dict[block_access["bn_var"]] = np.ones(bn_shape, dtype=np.float32)
                unfolded_state_dict[block_access["bn_beta"]] = folded_state_dict[block_access["bi"]]
                unfolded_state_dict[block_access["bn_gamma"]] = folded_state_dict[block_access["ls"]]
        return unfolded_state_dict

    def save_internal_states(self):
        np.savez(f"{self.args.results}/client_ID{self.id}_internal_states.npz", **self.internal_states)


    def load_internal_states(self):
        loaded_states = np.load(f"{self.args.results}/client_ID{self.id}_internal_states.npz",
                                allow_pickle=True)  # TODO get rid of allow_pickle
        self.internal_states = {k: loaded_states[k].item() for k in loaded_states.files}
        loaded_states.close()



'''
Copyright (c) 2023 Andrej Karpathy, MIT License
https://github.com/karpathy/llama2.c

Taken from llama code and lightly modified
Copyright (c) Meta Platforms, Inc. and affiliates.
This software may be used and distributed according to the terms of the Llama 2 Community License Agreement.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

The copyright in *this* software is being made available under the Clear BSD
License, included below. No patent rights, trademark rights and/or
other Intellectual Property Rights other than the copyrights concerning
the Software are granted under this license.

The Clear BSD License

Copyright (c) 2019-2025, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The NNCodec Authors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted (subject to the limitations in the disclaimer below) provided that
the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

     * Neither the name of the copyright holder nor the names of its
     contributors may be used to endorse or promote products derived from this
     software without specific prior written permission.

NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
'''
import os
import struct
import argparse
from typing import List

from sentencepiece import SentencePieceProcessor

TOKENIZER_MODEL = "tokenizer.model" # the llama sentencepiece tokenizer model

class Tokenizer:
    def __init__(self, tokenizer_model=None):
        model_path = tokenizer_model if tokenizer_model else TOKENIZER_MODEL
        assert os.path.isfile(model_path), model_path
        self.sp_model = SentencePieceProcessor(model_file=model_path)
        self.model_path = model_path

        # BOS / EOS token IDs
        self.n_words: int = self.sp_model.vocab_size()
        self.bos_id: int = self.sp_model.bos_id()
        self.eos_id: int = self.sp_model.eos_id()
        self.pad_id: int = self.sp_model.pad_id()
        #print(f"#words: {self.n_words} - BOS ID: {self.bos_id} - EOS ID: {self.eos_id}")
        assert self.sp_model.vocab_size() == self.sp_model.get_piece_size()

    def encode(self, s: str, bos: bool, eos: bool) -> List[int]:
        assert type(s) is str

        # t = self.sp_model.encode_as_pieces(s)
        # print("Tokenized Pieces:", t)
        # print("Piece IDs:", [self.sp_model.piece_to_id(piece) for piece in t])


        t = self.sp_model.encode(s)
        if bos:
            t = [self.bos_id] + t
        if eos:
            t = t + [self.eos_id]
        return t

    def decode(self, t: List[int]) -> str:
        return self.sp_model.decode(t)

    def export(self):

        # get all the tokens (postprocessed) and their scores as floats
        tokens, scores = [], []
        for i in range(self.n_words):

            # decode the token and light postprocessing
            t = self.sp_model.id_to_piece(i)
            s = self.sp_model.get_score(i)
            if i == self.bos_id:
                t = '\n<s>\n'
            elif i == self.eos_id:
                t = '\n</s>\n'
            t = t.replace('▁', ' ') # sentencepiece uses this character as whitespace
            b = t.encode('utf-8') # bytes of this token, utf-8 encoded

            tokens.append(b)
            scores.append(s)

        # record the max token length
        max_token_length = max(len(t) for t in tokens)

        # write to a binary file
        # the tokenizer.bin file is the same as .model file, but .bin
        tokenizer_bin = self.model_path.replace('.model', '.bin')
        with open(tokenizer_bin, 'wb') as f:
            f.write(struct.pack("I", max_token_length))
            for bytes, score in zip(tokens, scores):
                f.write(struct.pack("fI", score, len(bytes)))
                f.write(bytes)

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-t", "--tokenizer-model", type=str, help="optional path to custom tokenizer ")
    args = parser.parse_args()

    t = Tokenizer(args.tokenizer_model)
    t.export()



'''
The copyright in this software is being made available under the Clear BSD
License, included below. No patent rights, trademark rights and/or 
other Intellectual Property Rights other than the copyrights concerning 
the Software are granted under this license.

The Clear BSD License

Copyright (c) 2019-2025, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The NNCodec Authors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted (subject to the limitations in the disclaimer below) provided that
the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

     * Neither the name of the copyright holder nor the names of its
     contributors may be used to endorse or promote products derived from this
     software without specific prior written permission.

NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
'''
import numpy as np
import random
import torch
# import tensorflow as tf
from torch.utils.data import dataloader, random_split
from nncodec.framework.applications.datasets import imagenet200, imagenet1000, cifar10, cifar100, voc, V2X
from nncodec.framework.applications.utils import evaluation, train, transforms

def seed_worker(worker_id):
    worker_seed = torch.initial_seed() % 2**32
    np.random.seed(worker_seed)
    random.seed(worker_seed)

g = torch.Generator()
g.manual_seed(0)


class ModelSetting:

    def __init__(self, model_transform, evaluate, train, dataset, criterion):

        self.model_transform = model_transform
        self.evaluate = evaluate
        self.train = train
        self.dataset = dataset
        self.criterion = criterion
        self.train_cf = False

    def init_training(self, dataset_path, batch_size, num_workers):
        train_set = self.dataset(
            root=dataset_path,
            split='train'
        )
        train_loader = torch.utils.data.DataLoader(
                    train_set,
                    batch_size=batch_size,
                    shuffle=True,
                    num_workers=num_workers,
                    pin_memory=True,
                    collate_fn=getattr(train_set, "collate_fn", dataloader.default_collate),
                    sampler=getattr(train_set, "sampler", None),
                    worker_init_fn=seed_worker,
                    generator=g,
                )

        return train_loader

    def init_test(self, dataset_path, batch_size, num_workers):
        test_set = self.dataset(
            root=dataset_path,
            split='test'
        )

        test_loader = torch.utils.data.DataLoader(
            test_set,
            batch_size=batch_size,
            shuffle=False,
            num_workers=num_workers,
            pin_memory=True,
            collate_fn=getattr(test_set, "collate_fn", dataloader.default_collate),
            sampler=getattr(test_set, "sampler", None),
        )
        return test_set, test_loader

    
    def init_validation(self, dataset_path, batch_size, num_workers):
        val_set = self.dataset(
            root=dataset_path,
            split='val'
        )

        val_loader = torch.utils.data.DataLoader(
            val_set,
            batch_size=batch_size,
            shuffle=False,
            num_workers=num_workers,
            pin_memory=True,
            collate_fn=getattr(val_set, "collate_fn", dataloader.default_collate),
            sampler=getattr(val_set, "sampler", None),
            worker_init_fn=seed_worker,
            generator=g,
        )

        return val_set, val_loader

    def init_test_tef(self, dataset_path, batch_size, num_workers, model_name):
        test_set = self.dataset(
            root=dataset_path,
            split='test'
        )
        self.__model_name = model_name

        test_images, test_labels = zip(*test_set.imgs)
        test_loader = tf.data.Dataset.from_tensor_slices((list(test_images), list(test_labels)))
        test_loader = test_loader.map(lambda image, label: tf.py_function(self.tef_preprocess,
                                                                          inp=[image, label],
                                                                          Tout=[tf.float32, tf.int32]), num_parallel_calls=num_workers).batch(
                                                                                batch_size)

        return test_set, test_loader

    def init_validation_tef(self, dataset_path, batch_size, num_workers, model_name):
        val_set = self.dataset(
            root=dataset_path,
            split='val'
        )
        self.__model_name = model_name

        val_images, val_labels = zip(*val_set.imgs)
        val_loader = tf.data.Dataset.from_tensor_slices((list(val_images), list(val_labels)))
        val_loader = val_loader.map(lambda image, label: tf.py_function(self.tef_preprocess,
                                                                        inp=[image, label],
                                                                        Tout=[tf.float32, tf.int32]), num_parallel_calls=num_workers).batch(
                                                                                batch_size)

        return val_set, val_loader

    
    def tef_preprocess(
                    self,
                    image,
                    label
    ):
        image_size = 224

        if self.__model_name == 'EfficientNetB1':
            image_size = 240
        elif self.__model_name == 'EfficientNetB2':
            image_size = 260
        elif self.__model_name == 'EfficientNetB3':
            image_size = 300
        elif self.__model_name == 'EfficientNetB4':
            image_size = 380
        elif self.__model_name == 'EfficientNetB5':
            image_size = 456
        elif self.__model_name == 'EfficientNetB6':
            image_size = 528
        elif self.__model_name == 'EfficientNetB7':
            image_size = 600

        ##TODO make sure that this is a data transform only, i.e., is in use_case_init/__init__.py DATA_TRAFOS
        image, label = self.model_transform(image, label, image_size=image_size)

        if 'DenseNet' in self.__model_name:
            return tf.keras.applications.densenet.preprocess_input(image), label
        elif 'EfficientNet' in self.__model_name:
            return tf.keras.applications.efficientnet.preprocess_input(image), label
        elif self.__model_name == 'InceptionResNetV2':
            return tf.keras.applications.inception_resnet_v2.preprocess_input(image), label
        elif self.__model_name == 'InceptionV3':
            return tf.keras.applications.inception_v3.preprocess_input(image), label
        elif self.__model_name == 'MobileNet':
            return tf.keras.applications.mobilenet.preprocess_input(image), label
        elif self.__model_name == 'MobileNetV2':
            return tf.keras.applications.mobilenet_v2.preprocess_input(image), label
        elif 'NASNet' in self.__model_name:
            return tf.keras.applications.nasnet.preprocess_input(image), label
        elif 'ResNet' in self.__model_name and 'V2' not in self.__model_name:
            return tf.keras.applications.resnet.preprocess_input(image), label
        elif 'ResNet' in self.__model_name and 'V2' in self.__model_name:
            return tf.keras.applications.resnet_v2.preprocess_input(image), label
        elif self.__model_name == 'VGG16':
            return tf.keras.applications.vgg16.preprocess_input(image), label
        elif self.__model_name == 'VGG19':
            return tf.keras.applications.vgg19.preprocess_input(image), label
        elif self.__model_name == 'Xception':
            return tf.keras.applications.xception.preprocess_input(image), label


# supported use cases
use_cases = {

    "NNR_PYT_Telko": ModelSetting(None,
                                evaluation.evaluate_language_model,
                                train.train_language_model,
                                V2X,
                                None,
                                ),

    "NNR_PYT_VOC": ModelSetting(None,  # transforms.model_transform_ImageNet_to_CIFAR100,
                                evaluation.evaluate_classification_model,
                                train.train_classification_model,
                                voc,
                                torch.nn.CrossEntropyLoss()
                                ),

    "NNR_PYT_CIF10": ModelSetting(None,
                                  evaluation.evaluate_classification_model,
                                  train.train_classification_model,
                                  cifar10,
                                  torch.nn.CrossEntropyLoss()
                                  ),

    "NNR_PYT_CIF100":  ModelSetting(None,
                                    evaluation.evaluate_classification_model,
                                    train.train_classification_model,
                                    cifar100,
                                    torch.nn.CrossEntropyLoss()
                                    ),

    "NNR_PYT_IN200": ModelSetting(None,
                              evaluation.evaluate_classification_model,
                              train.train_classification_model,
                              imagenet200,
                              torch.nn.CrossEntropyLoss()
                              ),

    "NNR_PYT": ModelSetting(None,
                            evaluation.evaluate_classification_model,
                            train.train_classification_model,
                            imagenet1000,
                            torch.nn.CrossEntropyLoss()
                            ),

    "NNR_TEF": ModelSetting(transforms.transforms_tef_model_zoo,
                            evaluation.evaluate_classification_model_TEF,
                            None,
                            imagenet1000,
                            torch.nn.CrossEntropyLoss
                            )
}


'''
The copyright in this software is being made available under the Clear BSD
License, included below. No patent rights, trademark rights and/or 
other Intellectual Property Rights other than the copyrights concerning 
the Software are granted under this license.

The Clear BSD License

Copyright (c) 2019-2025, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The NNCodec Authors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted (subject to the limitations in the disclaimer below) provided that
the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

     * Neither the name of the copyright holder nor the names of its
     contributors may be used to endorse or promote products derived from this
     software without specific prior written permission.

NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
'''
import copy
import sys
from collections import OrderedDict
import numpy as np
from . import baseline
from . import codebook
from . import integer
from nncodec import nnc_core
from nncodec.nnc_core.nnr_model import NNRModelAccess, NNRBlockAccess, W_TYPES
from nncodec.nnc_core import hls
from nncodec.nnc_core.common import get_qp_from_stepsize
from timeit import default_timer as timer

def __print_output_line( outputString, verbose=True ):
    if verbose:
        sys.stdout.write(outputString)
        sys.stdout.flush()

def del_param(approx_data, approx_info, param):
    del approx_data["parameters"][param]
    approx_data["scan_order"].pop(param, None)
    approx_info.get("qp", {}).pop(param, None)
    approx_info.get("dq_flag", {}).pop(param, None)
    
def init_approx_data(
    parameters,
    model_info,
    qp_density,
    scan_order,
): 
    approx_data = {
        "approx_method": {},
        "qp_density": np.int32(qp_density),
        "qp": {},
        "dq_flag": {},
        "decomposition_rank": {},
        "g_number_of_rows": {},
        "scan_order": {},
        "parameters": copy.copy(parameters),
        "compressed_parameter_types": {},
        "codebooks": {},
        "codebooks_egk": {},
        "codebook_zero_offsets": {},
    }
    for x in parameters:
        assert (x.endswith("_G") or x.endswith("_H")) == (("_G" in x) or ("_H" in x))
        if x.endswith("_G") or x.endswith("_H"):
            if len(model_info["parameter_dimensions"][x[:-2]]) > 1:
                approx_data["scan_order"][x] = np.int32(scan_order) 
        elif len(model_info["parameter_dimensions"][x]) > 1:
            approx_data["scan_order"][x] = np.int32(scan_order) 
        else:
            continue

    for block_id in model_info["block_identifier"].values():
        if block_id != None:
            block_access = NNRBlockAccess(model_info, block_id)
            cpt = 0
            if block_access.bn_gamma:
                cpt += hls.BlockParameterTypes.NNR_CPT_BN
            if block_access.bi in approx_data["parameters"]: 
                cpt += hls.BlockParameterTypes.NNR_CPT_BI
            if block_access.dc_g in approx_data["parameters"]: 
                cpt += hls.BlockParameterTypes.NNR_CPT_DC
                par_dc_g = block_access.dc_g
                #par_dc_h = block_access.dc_h
                approx_data["decomposition_rank"][block_id] = approx_data["parameters"][par_dc_g].shape[1] 
                approx_data["g_number_of_rows"][block_id] = approx_data["parameters"][par_dc_g].shape[0] 
            if block_access.ls in approx_data["parameters"]: 
                cpt += hls.BlockParameterTypes.NNR_CPT_LS
            approx_data["compressed_parameter_types"][block_id] = cpt

    return approx_data

        
def fold_bn(model_info, approx_data, ap_info, bnf_mapping=False):
    model_access = NNRModelAccess(model_info)
    if bnf_mapping:
        model_info["bnf_map"] = OrderedDict({k: k for k in approx_data["parameters"].keys()})
        model_info["mdl_blocks"] = {}
    for block_access in model_access.blocks_and_params():
        block_id = block_access.block_id
        if block_id is None:
            continue
        cpt = approx_data["compressed_parameter_types"][block_id]
        ad = approx_data["parameters"]
        assert not approx_data["approx_method"]
        eps = 1e-3 if model_info['topology_storage_format'] == nnc_core.nnr_model.TopologyStorageFormat.NNR_TPL_TEF else 1e-5
        if cpt & hls.BlockParameterTypes.NNR_CPT_BN != 0:
            delta = block_access.bi
            bn_shape = ad[block_access.bn_mean].shape
            dq_flag = ap_info.approx_info["dq_flag"][block_access.bn_mean]
            assert (cpt & hls.BlockParameterTypes.NNR_CPT_BI == 0) == (delta not in ad)
            if cpt & hls.BlockParameterTypes.NNR_CPT_BI == 0:
                ad[delta] = np.zeros(bn_shape, dtype=np.float32)
                approx_data["compressed_parameter_types"][block_id] += hls.BlockParameterTypes.NNR_CPT_BI
                if ap_info.approx_info["approx_method"] == "uniform":
                    ap_info.approx_info["qp"][delta] = ap_info.qp_other
                    ap_info.approx_info["dq_flag"][delta] = dq_flag
            alpha = block_access.ls
            assert (cpt & hls.BlockParameterTypes.NNR_CPT_LS == 0) == (alpha not in ad)
            if cpt & hls.BlockParameterTypes.NNR_CPT_LS == 0:
                assert bn_shape == ad[block_access.bn_mean].shape
                ad[alpha] = np.ones(bn_shape, dtype=np.float32)
                approx_data["compressed_parameter_types"][block_id] += hls.BlockParameterTypes.NNR_CPT_LS
                if ap_info.approx_info["approx_method"] == "uniform":
                    ap_info.approx_info["qp"][alpha] = ap_info.qp_lsa
                    ap_info.approx_info["dq_flag"][alpha] = dq_flag

            g = ad[block_access.bn_gamma] / np.sqrt( ad[block_access.bn_var] + eps )
            del_param(approx_data, ap_info.approx_info, block_access.bn_gamma)
            del_param(approx_data, ap_info.approx_info, block_access.bn_var)
            ad[alpha] *= g
            ad[delta] = (ad[delta] - ad[block_access.bn_mean]) * g + ad[block_access.bn_beta]
            del_param(approx_data, ap_info.approx_info, block_access.bn_mean)
            del_param(approx_data, ap_info.approx_info, block_access.bn_beta)
            approx_data["compressed_parameter_types"][block_id] -= hls.BlockParameterTypes.NNR_CPT_BN

            if bnf_mapping:
                model_info["mdl_blocks"][block_id] = {k: getattr(block_access, k) for k in dir(block_access)
                                                      if isinstance(getattr(block_access, k), str) and not k.startswith('_')}
                del model_info["bnf_map"][block_access.bn_var]
                del model_info["bnf_map"][block_access.bn_mean]
                model_info["bnf_map"][block_access.bn_beta] = delta
                model_info["bnf_map"][block_access.bn_gamma] = alpha

        elif bnf_mapping:
            model_info["mdl_blocks"][block_id] = {k: getattr(block_access, k) for k in dir(block_access)
                                                  if isinstance(getattr(block_access, k), str) and not k.startswith('_')}
    if bnf_mapping:
        model_info["bnf_map"] = {model_info["bnf_map"][k]: k for k in model_info["parameter_type"]
                                  if k in model_info["bnf_map"]} # re-ordering
        return model_info

            
def unfold_bn(model_info, approx_data):
    model_access = NNRModelAccess(model_info)
    ad = approx_data["parameters"]
    for block_access in model_access.blocks_and_params():
        block_id = block_access.block_id
        if block_id is None:
            continue

        bn_absent = approx_data["compressed_parameter_types"][block_id] & hls.BlockParameterTypes.NNR_CPT_BN == 0
        bn_folded = bn_absent and (block_access.bn_gamma in model_info["parameter_type"])
        if bn_folded:
            approx_data["compressed_parameter_types"][block_id] += hls.BlockParameterTypes.NNR_CPT_BN
            #ad = approx_data["parameters"]

            delta = block_access.bi
            dims = approx_data["parameters"][delta].shape
            if delta not in model_info["parameter_type"]:
                assert approx_data["compressed_parameter_types"][block_id] & hls.BlockParameterTypes.NNR_CPT_BI != 0
                approx_data["parameters"][block_access.bn_beta] = approx_data["parameters"][delta]
                del approx_data["parameters"][delta]
                approx_data["compressed_parameter_types"][block_id] -= hls.BlockParameterTypes.NNR_CPT_BI
            else:
                approx_data["parameters"][block_access.bn_beta] = np.zeros(dims, dtype=np.float32)

            approx_data["parameters"][block_access.bn_mean]  = np.zeros(dims, dtype=np.float32)
            approx_data["parameters"][block_access.bn_gamma] = np.ones(dims, dtype=np.float32)
            approx_data["parameters"][block_access.bn_var]   = np.ones(dims, dtype=np.float32)


def set_lsa(model_info, approx_data, lsa_params):
    for k, v in lsa_params.items():
        approx_data["parameters"][k] = v.reshape([v.shape[0]])
        bi = model_info["block_identifier"].get(k, None)
        if bi is not None:
            approx_data["compressed_parameter_types"][bi] |= hls.BlockParameterTypes.NNR_CPT_LS

def apply_lsa(model_info, approx_data):
    assert not approx_data["approx_method"]
    model_access = NNRModelAccess(model_info)
    for block_access in model_access.blocks_and_params():
        block_id = block_access.block_id
        if block_id is None:
            continue
        cpt = approx_data["compressed_parameter_types"][block_id]
        if cpt & hls.BlockParameterTypes.NNR_CPT_LS != 0:
            ls = approx_data['parameters'].pop(block_access.ls)
            _ = model_info["parameter_index"].pop(block_access.ls, None)
            _ = model_info["block_identifier"].pop(block_access.ls, None)

            if cpt & hls.BlockParameterTypes.NNR_CPT_DC != 0:
                w = approx_data['parameters'][block_access.dc_g]
            else:
                w = approx_data['parameters'][block_access.w]
            dims_ls = [-1] + [1] * (w.ndim - 1)
            w *= ls.reshape(dims_ls)
            approx_data['compressed_parameter_types'][block_id] -= hls.BlockParameterTypes.NNR_CPT_LS

def recompose_params(model_info, approx_data_in):
    assert not approx_data_in["approx_method"]
    approx_data_out = {k: copy.copy(v) for k, v in approx_data_in.items()} # create copies of dicts in approx_data

    model_access = NNRModelAccess(model_info)
    for block_access in model_access.blocks_and_params():
        block_id = block_access.block_id
        if block_id is None:
            continue
        cpt = approx_data_out["compressed_parameter_types"][block_id]
        if cpt & hls.BlockParameterTypes.NNR_CPT_DC != 0:
            g = approx_data_out["parameters"].pop(block_access.dc_g)
            h = approx_data_out["parameters"].pop(block_access.dc_h)

            recomposed_w = g.dot(h)
            recomposed_w = recomposed_w.reshape(model_info["parameter_dimensions"][block_access.w])

            approx_data_out["parameters"][block_access.w] = recomposed_w
            approx_data_out['compressed_parameter_types'][block_id] -= hls.BlockParameterTypes.NNR_CPT_DC

            param_id_g = model_info["parameter_index"][block_access.dc_g].pop()
            model_info["parameter_index"][block_access.w] = param_id_g
            del(model_info["block_identifier"][block_access.dc_g])
            del(model_info["parameter_index"][block_access.dc_h])
            del(model_info["block_identifier"][block_access.dc_h])

    resorted_param_dict = dict()
    resorted_param_id_dict = {k: v for k, v in sorted(model_info["parameter_index"].items(), key=lambda item: item[1])}
    for param in resorted_param_id_dict.keys():
        resorted_param_dict[param] = copy.deepcopy(approx_data_out["parameters"][param])

    approx_data_out["parameters"] = resorted_param_dict

    return approx_data_out


def inference_based_qp_opt( 
        approx_info,
        model_info,
        model_executer,
        approx_data,
        enc_info,
        verbose,
    ):
    approx_data_qp = approx(
        approx_info,
        model_info,
        approx_data,
        enc_info,
    )
    rec_approx_data_qp = copy.deepcopy(approx_data_qp)
    rec(
        rec_approx_data_qp,
    )

    ##encode
    start = timer()
    __print_output_line("\tIOQ: PROCESSING QP FOR ALL TENSORS...", verbose=verbose)
    bitstream_qp, oob = nnc_core.coder.encode(enc_info, model_info, approx_data_qp)

    acc_qp = model_executer.eval_model(
        rec_approx_data_qp["parameters"],
        False,
    )

    refBSSize = len(bitstream_qp)
    refAcc = list(acc_qp.values())[0]

    bestCost = 0.0
    end = timer()
    __print_output_line("DONE in {:.4f} s\n".format( end-start ), verbose=verbose)


    ############################ eval with QP-1
    start = timer()
    __print_output_line("\tIOQ: PROCESSING QP-1 FOR ALL TENSORS...", verbose=verbose)

    approx_info_qp = copy.deepcopy(approx_info)

    for p in approx_info_qp["qp"].keys():
        if model_info["parameter_type"][p] in nnc_core.nnr_model.W_TYPES:
            approx_info_qp["qp"][p] -= 1

    approx_data_qp = approx(
        approx_info_qp,
        model_info,
        approx_data,
        enc_info,
    )
    rec_approx_data_qp = copy.deepcopy(approx_data_qp)
    rec(
        rec_approx_data_qp,
    )

    ##encode
    bitstream_qp, oob = nnc_core.coder.encode(enc_info, model_info, approx_data_qp)
    ##eval
    acc_qp = model_executer.eval_model(
        rec_approx_data_qp["parameters"],
        False,
    )

    currBSSize = len(bitstream_qp)
    currAcc = list(acc_qp.values())[0]

    diffBR = currBSSize - refBSSize
    diffAcc = refAcc - currAcc

    lambdaM1 = -diffAcc/diffBR

    end = timer()
    __print_output_line("DONE in {:.4f} s\n".format( end-start ), verbose=verbose)


    ############################### eval with QP+1
    start = timer()
    __print_output_line("\tIOQ: PROCESSING QP+1 FOR ALL TENSORS...", verbose=verbose)

    approx_info_qp = copy.deepcopy(approx_info)

    for p in approx_info_qp["qp"].keys():
        if model_info["parameter_type"][p] in nnc_core.nnr_model.W_TYPES:
            approx_info_qp["qp"][p] += 1

    approx_data_qp = approx(
        approx_info_qp,
        model_info,
        approx_data,
        enc_info,
    )
    rec_approx_data_qp = copy.deepcopy(approx_data_qp)
    rec(
        rec_approx_data_qp,
    )

    ##encode
    bitstream_qp, oob = nnc_core.coder.encode(enc_info, model_info, approx_data_qp)
    ##eval
    acc_qp = model_executer.eval_model(
        rec_approx_data_qp["parameters"],
        False,
    )

    currBSSize = len(bitstream_qp)
    currAcc = list(acc_qp.values())[0]

    diffBR = currBSSize - refBSSize
    diffAcc = refAcc - currAcc

    lambdaP1 = -diffAcc/diffBR

    end = timer()
    __print_output_line("DONE in {:.4f} s\n".format( end-start ), verbose=verbose)

    ################################

    ##sort parameters by size
    mapParamToSize = []
    approx_info_qp = copy.deepcopy(approx_info)
    for p in rec_approx_data_qp["parameters"]:
        if model_info["parameter_type"][p] in nnc_core.nnr_model.W_TYPES:
            mapParamToSize.append([p , np.size(approx_data_qp["parameters"][p])])

    mapParamToSize.sort(key = lambda x: x[1],reverse=True)

    setNeg = [-4, -3, -2, -1 ]
    setPos = [1, 2, 3, 4 ]

    qpOffsetSet = [setNeg, setPos]

    timeLastQp = "n/a"

    for iParam, item in enumerate(mapParamToSize[1::]):
        for iQpSet, qpSet in enumerate(qpOffsetSet):
            for iQpOff, qp_off in enumerate(qpSet):
                __print_output_line("\r\tIOQ: PROCESSING TENSOR {}/{} AND QP {}/{} (LAST QP-ITERATION TOOK: {})".format( iParam+1, len(mapParamToSize)-1, iQpSet*len(qpOffsetSet[0])+ iQpOff + 1, len(qpOffsetSet[0])+len(qpOffsetSet[1]), timeLastQp), verbose=verbose)
                start = timer()
                approx_info_qp_curr = copy.deepcopy(approx_info_qp)
                approx_info_qp_curr["qp"][item[0]] = approx_info["qp"][item[0]] + qp_off

                approx_data_qp = approx(
                    approx_info_qp_curr,
                    model_info,
                    approx_data,
                    enc_info,
                )

                rec_approx_data_qp = copy.deepcopy(approx_data_qp)
                rec(
                    rec_approx_data_qp,
                )

                ##encode
                bitstream_qp, oob = nnc_core.coder.encode(enc_info, model_info, approx_data_qp)

                ##eval
                acc_qp = model_executer.eval_model(
                    rec_approx_data_qp["parameters"],
                    False,
                )

                currBSSize = len(bitstream_qp)
                currAcc = list(acc_qp.values())[0]

                diffBR = currBSSize - refBSSize
                diffAcc = refAcc - currAcc

                lamb = max( (lambdaP1 + lambdaM1) / 2, 0.0 )

                currCost = diffAcc + lamb * diffBR

                if currCost < bestCost:
                    approx_info_qp = copy.deepcopy(approx_info_qp_curr)
                    bestCost = currCost
                
                end = timer()
                timeLastQp = "{:.4f} s".format( end-start )

    __print_output_line("\n")
    approx_info.clear()
    approx_info.update(approx_info_qp)


def run_ft_and_lsa(model_info, approx_data, ap_info, model_executer, block_id_and_param_type, lsa_flag, ft_flag, use_dq, verbose, wandb_logging):
    approx_info_ft = copy.deepcopy(ap_info.approx_info)
    if not lsa_flag:
        approx_info_ft["to_approximate"] = W_TYPES
    else:
        approx_info_ft["to_approximate"].remove('weight.ls')
    approx_data_ft = nnc_core.approximator.approx(approx_info_ft, model_info, approx_data)
    nnc_core.approximator.rec(approx_data_ft)

    tuned_params = model_executer.tune_model(
        parameters=approx_data_ft['parameters'],
        param_types=model_info['parameter_type'],
        lsa_flag=lsa_flag,
        ft_flag=ft_flag,
        verbose=verbose,
        wandb_logging=wandb_logging

    )
    
    lsa_params = tuned_params[0]
    ft_params  = tuned_params[1]

    if ft_flag:
        approx_data["parameters"].update(ft_params)
    if lsa_flag:
        if block_id_and_param_type:
            nnc_core.approximator.set_lsa(model_info, approx_data, lsa_params)
            nnc_core.nnr_model.add_lsa_to_block_id_and_param_type( block_id_and_param_type, lsa_params )
        else:
            approx_data["parameters"].update(lsa_params)
        ap_info.set_ls_qps(model_info, approx_data, 1 if use_dq else 0)


def approx(approx_info, model_info, approx_data, enc_info=None):

    approx_method = approx_info['approx_method']
    
    approx_data = integer.skip_approx( approx_info, model_info, approx_data )

    if approx_method == 'codebook':
        approx_data, approx_info = codebook.approx(approx_info, model_info, approx_data, enc_info=enc_info)

    return baseline.approx(approx_info, model_info, approx_data, enc_info=enc_info)
    

def rec(approx_data):
    for param in approx_data['parameters']:
        if param in approx_data["approx_method"]:
            if approx_data["approx_method"][param] == 'uniform':
                baseline.rec(param, approx_data)
            elif approx_data["approx_method"][param] == 'codebook':
                codebook.rec(param, approx_data)
            elif approx_data["approx_method"][param] == 'skip':
                integer.skip_rec(param, approx_data)
            else:
                assert param not in approx_data["approx_method"], "unknown approx_method"

def check_array_all_zero_or_scalar(x):
    if np.isscalar(x):
        return isinstance(x, (int, np.integer))
    elif isinstance(x, np.ndarray) and x.ndim == 0:
        return isinstance(x.item(), (int, np.integer))
    elif isinstance(x, np.ndarray) and np.all(x == 0):
        return True
    else:
        return False

class ApproxInfo():
    def __init__(
        self,
        approx_data,
        model_info,
        approx_method,
        codebook_mode,
        qp,
        opt_qp,
        disable_dq,
        cabac_unary_length_minus1,
        lambda_scale,
        nonweight_qp=None,
        qp_per_tensor=None,
        int_quant_bw=False
    ):
        self.__approx_info = {
            "approx_method": "codebook" if codebook_mode > 0 else approx_method,
            "codebook_mode": codebook_mode,
            "dq_flag": {x: 0 if disable_dq else 1 for x in approx_data["parameters"]},
            "lambda_scale": lambda_scale,
            "cabac_unary_length_minus1": cabac_unary_length_minus1,
            "to_approximate": nnc_core.nnr_model.W_TYPES + nnc_core.nnr_model.O_TYPES,
        }

        if approx_method == "uniform" or approx_method == "codebook":
            qp = np.int32(qp)
            qp_density = approx_data["qp_density"]
            self.__qp_other = nonweight_qp if nonweight_qp else qp - (2 << qp_density)  # same as dividing the stepsize by 4
            self.__qp_lsa = nonweight_qp if nonweight_qp else qp - (2 << qp_density)#qp - (8 << qp_density)
            self.approx_info["qp"] = {}
            if not int_quant_bw:
                for x in approx_data["parameters"]:
                    if x not in model_info["parameter_index"] and (x.endswith("_G") or x.endswith("_H")):
                        assert model_info["parameter_type"][x[:-2]] in nnc_core.nnr_model.W_TYPES, "Unexpected."
                        self.approx_info["qp"][x] = qp
                    else:
                        self.approx_info["qp"][x] = qp if model_info["parameter_type"][x] in nnc_core.nnr_model.W_TYPES else self.qp_other
                if qp_per_tensor is not None:
                    assert type(qp_per_tensor) is dict, "qp_per_tensor must be a dict!"
                    for x in approx_data["parameters"]:
                        self.approx_info["qp"][x] = qp_per_tensor.get(x, self.approx_info["qp"][x])
                if opt_qp:
                    self._modify_qp(approx_data, model_info)
            else:
                uint_support = True  # allows unsigned integer types, if params contain only positive values
                self.approx_info["integer_aligned_bitdepth"] = int_quant_bw
                self.approx_info["unsigned_integer_support"] = uint_support
                for x in approx_data["parameters"]:
                    if check_array_all_zero_or_scalar(approx_data["parameters"][x]):#".num_batches_tracked" in x or np.sum(approx_data["parameters"][x]) == 0:
                        self.approx_info["qp"][x] = np.int32(0)
                    else:
                        if int_quant_bw > 7 or model_info["parameter_type"][x] in W_TYPES:
                            bw = int_quant_bw
                        else: # non-weight params at least in 16bit
                            bw = 8
                            print(f"INFO: Since {x} seems to be no weight tensor, it was quantized to {bw} bit instead of {int_quant_bw} bit.")
                        self.approx_info["qp"][x] = self._get_intaligned_qp(param=approx_data["parameters"][x],
                                                                            qp_density=qp_density,
                                                                            bw=bw,
                                                                            uint_support=uint_support)


    @property
    def qp_lsa(self):
        return self.__qp_lsa

    @property
    def qp_other(self):
        return self.__qp_other

    @property
    def approx_info(self):
        return self.__approx_info

    def apply_qp(self, approx_data, model_info, qp, nonweight_qp=None):
        qp = np.int32(qp)
        qp_density = approx_data["qp_density"]
        self.__qp_other = nonweight_qp if nonweight_qp else qp - (2 << qp_density)
        self.__qp_lsa = nonweight_qp if nonweight_qp else qp - (2 << qp_density)#qp - (8 << qp_density)
        self.approx_info["qp"] = {}
        for x in approx_data["parameters"]:
            if x not in model_info["parameter_index"] and (x.endswith("_G") or x.endswith("_H")):
                assert model_info["parameter_type"][x[:-2]] in nnc_core.nnr_model.W_TYPES, "Unexpected."
                self.approx_info["qp"][x] = qp
            else:
                if model_info["parameter_type"][x] in nnc_core.nnr_model.W_TYPES:
                    self.approx_info["qp"][x] = qp
                else:
                    self.approx_info["qp"][x] = self.qp_other

    def _get_intaligned_qp(self, param, qp_density, bw, uint_support):
        if uint_support and len(param[param < 0]) == 0:
            denom = 2 ** bw - 1
        else:
            denom = 2 ** (bw - 1) - 1
        step_size = np.percentile(abs(param), 99.95) / denom
        if step_size == 0:
            step_size = np.max(abs(param)) / denom
        qp = get_qp_from_stepsize(step_size, qp_density)
        max_qp = 2 ** (6 + qp_density - 1)
        if qp < -max_qp or qp > max_qp - 1:
            print(f"INFO QP {qp} clipped to [{-max_qp},{max_qp - 1}]")
            qp = qp.clip(-max_qp, max_qp - 1)
        return np.int32(np.round(qp))

    def _modify_qp(self, approx_data, model_info):
        param_types = ["weight"]#["fc.weight", "conv.weight"]
        param_names = []
        param_sizes = []
        param_std = []
        for k, v in approx_data["parameters"].items():
            param_w = k[:-2] if ( k.endswith("_G") or k.endswith("_H") ) else k 
            if model_info["parameter_type"][param_w] in param_types:
                if ( k.endswith("_G") or k.endswith("_H") ):
                    if k.endswith("_G") : continue
                    assert k.endswith("_H")
                    g = approx_data["parameters"][param_w + "_G"].shape
                    h = approx_data["parameters"][param_w + "_H"].shape
                    if len( h ) == 4:
                        assert h[0] == 1
                        assert h[1] == 1
                    assert h[-2] == g[-1]
                    s = np.prod(g[:-1]) * h[-1]
                    param_names.append(param_w + "_G")
                    param_sizes.append(0)
                    param_std.append(0)
                    param_names.append(param_w + "_H")
                    param_sizes.append(s)
                    param_std.append(np.std(np.concatenate(
                        (approx_data["parameters"][param_w + "_G"].flatten(),
                         approx_data["parameters"][param_w + "_H"].flatten()), axis=0)))
                else:
                    param_names.append(k)
                    param_sizes.append(v.size)
                    param_std.append(np.std(v))

        rel_layer_sizes = np.array(param_sizes) / sum(param_sizes)
        rel_layer_std = np.array(param_std) / max(param_std)

        shares = rel_layer_sizes + (.1 * (1 - rel_layer_std))

        w = dict(zip(param_names, shares))
        for name in param_names:
            qp = self.__approx_info['qp'][name]
            if w[name] > .5: w[name] = .15
            self.__approx_info['qp'][name] = np.int32(round(qp * (1 - w[name])))
            if name.endswith( "_H" ):
                self.__approx_info['qp'][name[:-2]+"_G"] = self.__approx_info['qp'][name]

    def set_ls_qps(self, model_info, approx_data, dq_flag):
        for block_access in NNRModelAccess(model_info).blocks_and_params():
            if block_access.block_id is not None:
                cpt = approx_data["compressed_parameter_types"][block_access.block_id]
                if cpt & hls.BlockParameterTypes.NNR_CPT_LS != 0:
                    self.approx_info["qp"][block_access.ls] = self.qp_lsa
                    self.approx_info["dq_flag"][block_access.ls] = dq_flag
            

